# @file
# Code to test UEFI MultiPhase module
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import unittest
from edk2toollib.uefi.uefi_multi_phase import (EfiVariableAttributes,
                                               EFI_VARIABLE_NON_VOLATILE, EFI_VARIABLE_RUNTIME_ACCESS,
                                               EFI_VARIABLE_BOOTSERVICE_ACCESS,
                                               EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS)


class TestUefiMultiphase (unittest.TestCase):

    def test_string_conversion(self):
        attributes = EfiVariableAttributes(EFI_VARIABLE_NON_VOLATILE
                                           | EFI_VARIABLE_RUNTIME_ACCESS | EFI_VARIABLE_BOOTSERVICE_ACCESS)
        string = str(attributes)

        self.assertTrue("EFI_VARIABLE_RUNTIME_ACCESS" in string)
        self.assertTrue("EFI_VARIABLE_NON_VOLATILE" in string)
        self.assertTrue("EFI_VARIABLE_BOOTSERVICE_ACCESS" in string)

    def test_empty(self):
        attributes = EfiVariableAttributes(0)

        self.assertEqual(str(attributes), "")
        self.assertEqual(int(attributes), 0)

    def test_int_to_alternate(self):

        attributes = EfiVariableAttributes(EFI_VARIABLE_NON_VOLATILE)
        self.assertEqual(str(attributes), "EFI_VARIABLE_NON_VOLATILE")
        self.assertEqual(attributes.get_short_string(), "NV")
        self.assertEqual(int(attributes), EFI_VARIABLE_NON_VOLATILE)

        attributes.update(EFI_VARIABLE_NON_VOLATILE | EFI_VARIABLE_BOOTSERVICE_ACCESS)
        self.assertEqual(str(attributes), "EFI_VARIABLE_BOOTSERVICE_ACCESS,EFI_VARIABLE_NON_VOLATILE")
        self.assertEqual(attributes.get_short_string(), "BS,NV")
        self.assertEqual(int(attributes), EFI_VARIABLE_NON_VOLATILE | EFI_VARIABLE_BOOTSERVICE_ACCESS)

        attributes.update(EFI_VARIABLE_NON_VOLATILE | EFI_VARIABLE_BOOTSERVICE_ACCESS
                          | EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS)
        self.assertEqual(
            str(attributes), "EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS,EFI_VARIABLE_BOOTSERVICE_ACCESS,EFI_VARIABLE_NON_VOLATILE")  # noqa
        self.assertEqual(attributes.get_short_string(), "AT,BS,NV")
        self.assertEqual(int(attributes), EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS
                         | EFI_VARIABLE_NON_VOLATILE | EFI_VARIABLE_BOOTSERVICE_ACCESS)

    def test_string_to_alternate(self):
        attributes = EfiVariableAttributes("EFI_VARIABLE_NON_VOLATILE")
        self.assertEqual(str(attributes), "EFI_VARIABLE_NON_VOLATILE")
        self.assertEqual(attributes.get_short_string(), "NV")
        self.assertEqual(int(attributes), EFI_VARIABLE_NON_VOLATILE)

        attributes.update("EFI_VARIABLE_BOOTSERVICE_ACCESS,EFI_VARIABLE_NON_VOLATILE")
        self.assertEqual(str(attributes), "EFI_VARIABLE_BOOTSERVICE_ACCESS,EFI_VARIABLE_NON_VOLATILE")
        self.assertEqual(attributes.get_short_string(), "BS,NV")
        self.assertEqual(int(attributes), EFI_VARIABLE_NON_VOLATILE | EFI_VARIABLE_BOOTSERVICE_ACCESS)

        attributes.update(
            "EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS,EFI_VARIABLE_BOOTSERVICE_ACCESS,EFI_VARIABLE_NON_VOLATILE")  # noqa
        self.assertEqual(
            str(attributes), "EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS,EFI_VARIABLE_BOOTSERVICE_ACCESS,EFI_VARIABLE_NON_VOLATILE")  # noqa
        self.assertEqual(attributes.get_short_string(), "AT,BS,NV")
        self.assertEqual(int(attributes), EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS
                         | EFI_VARIABLE_NON_VOLATILE | EFI_VARIABLE_BOOTSERVICE_ACCESS)

    def test_short_string_to_alternate(self):
        attributes = EfiVariableAttributes("NV")
        self.assertEqual(str(attributes), "EFI_VARIABLE_NON_VOLATILE")
        self.assertEqual(attributes.get_short_string(), "NV")
        self.assertEqual(int(attributes), EFI_VARIABLE_NON_VOLATILE)

        attributes.update("BS,NV")
        self.assertEqual(str(attributes), "EFI_VARIABLE_BOOTSERVICE_ACCESS,EFI_VARIABLE_NON_VOLATILE")
        self.assertEqual(attributes.get_short_string(), "BS,NV")
        self.assertEqual(int(attributes), EFI_VARIABLE_NON_VOLATILE | EFI_VARIABLE_BOOTSERVICE_ACCESS)

        attributes.update("AT,BS,NV")
        self.assertEqual(
            str(attributes), "EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS,EFI_VARIABLE_BOOTSERVICE_ACCESS,EFI_VARIABLE_NON_VOLATILE")  # noqa
        self.assertEqual(attributes.get_short_string(), "AT,BS,NV")
        self.assertEqual(int(attributes), EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS
                         | EFI_VARIABLE_NON_VOLATILE | EFI_VARIABLE_BOOTSERVICE_ACCESS)

    def test_with_spaces_to_alternate(self):
        attributes = EfiVariableAttributes("BS, NV")
        self.assertEqual(str(attributes), "EFI_VARIABLE_BOOTSERVICE_ACCESS,EFI_VARIABLE_NON_VOLATILE")
        self.assertEqual(attributes.get_short_string(), "BS,NV")
        self.assertEqual(int(attributes), EFI_VARIABLE_NON_VOLATILE | EFI_VARIABLE_BOOTSERVICE_ACCESS)

        attributes.update("EFI_VARIABLE_BOOTSERVICE_ACCESS,EFI_VARIABLE_NON_VOLATILE")
        self.assertEqual(str(attributes), "EFI_VARIABLE_BOOTSERVICE_ACCESS,EFI_VARIABLE_NON_VOLATILE")
        self.assertEqual(attributes.get_short_string(), "BS,NV")
        self.assertEqual(int(attributes), EFI_VARIABLE_NON_VOLATILE | EFI_VARIABLE_BOOTSERVICE_ACCESS)



# @file tpm2_defs_test.py
# This file contains utility classes to help interpret definitions from the
# Tpm20.h header file in TianoCore.
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import unittest
import pytest
from edk2toollib.uefi.fmp_capsule_header import FmpCapsuleHeaderClass


class TestFmpCapsuleHeaderClass(unittest.TestCase):
    @pytest.mark.skip(reason="test is incomplete")
    def test_should_successfully_decode_test_payload(self):
        pass

    def test_embedded_driver_count_should_track_additions(self):
        test_header = FmpCapsuleHeaderClass()
        self.assertEqual(test_header.EmbeddedDriverCount, 0)
        test_header.AddEmbeddedDriver(b'dummydriver')
        self.assertEqual(test_header.EmbeddedDriverCount, 1)
        test_header.AddEmbeddedDriver(b'dummydriver2')
        self.assertEqual(test_header.EmbeddedDriverCount, 2)

    def test_payload_item_count_should_track_additions(self):
        test_header = FmpCapsuleHeaderClass()
        self.assertEqual(test_header.PayloadItemCount, 0)
        test_header.AddFmpCapsuleImageHeader(b'dummyheader')
        self.assertEqual(test_header.PayloadItemCount, 1)
        test_header.AddFmpCapsuleImageHeader(b'dummyheader2')
        self.assertEqual(test_header.PayloadItemCount, 2)

    def test_encoding_twice_should_yield_identical_results(self):
        test_header = FmpCapsuleHeaderClass()
        test_header.AddEmbeddedDriver(b'dummydriver')
        encode_1 = test_header.Encode()
        encode_2 = test_header.Encode()
        self.assertEqual(encode_1, encode_2)



# @file path_utilities_test.py
# Contains unit test routines for the path_utilities class.
#
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import unittest
import os
import sys
import tempfile
import shutil
from edk2toollib.uefi.edk2.path_utilities import Edk2Path
from pathlib import Path, PurePath


class PathUtilitiesTest(unittest.TestCase):

    def _make_file_helper(self, parentpath: str, filename: str) -> None:
        ''' simple internal helper to make a file and write basic content'''
        with open(os.path.join(parentpath, filename), "w") as f:
            f.write("Unit test content")

    def _make_edk2_module_helper(self, parentpath: str, modname: str, extension_case_lower: bool = True) -> str:
        ''' simple internal helper to make a folder and inf like edk2
        return the path to the module folder
        '''
        modfolder = os.path.join(parentpath, modname)
        os.makedirs(modfolder, exist_ok=True)
        fn = modname + (".inf" if extension_case_lower else ".INF")
        self._make_file_helper(modfolder, fn)
        return modfolder

    def _make_edk2_package_helper(self, path: str, packagename: str, extension_case_lower: bool = True) -> str:
        ''' simple internal helper to make an ed2 package as follows:

        path/
          packagename/
            packagename.dec
            module1/
              module1.inf
            module2/
              module2.inf
              X64/
                TestFile.c

        return the path to the new package
        '''
        pkgfolder = os.path.join(path, packagename)
        os.makedirs(pkgfolder, exist_ok=True)
        pn = packagename + (".dec" if extension_case_lower else ".DEC")
        self._make_file_helper(pkgfolder, pn)
        self._make_edk2_module_helper(pkgfolder, "module1", extension_case_lower=extension_case_lower)
        p2 = self._make_edk2_module_helper(pkgfolder, "module2", extension_case_lower=extension_case_lower)
        p3 = os.path.join(p2, "X64")
        os.makedirs(p3)
        self._make_file_helper(p3, "TestFile.c")
        return pkgfolder

    def setUp(self):
        ''' unittest fixture run before each test.
        Create a tmpdir and set the current working dir there'''
        self.tmp = tempfile.mkdtemp()
        self.precwd = os.getcwd()
        os.chdir(self.tmp)  # move to tempfile root

    def tearDown(self):
        ''' unittest fixture run after each test.
        Delete the tempdir and set the current working dir back'''
        os.chdir(self.precwd)
        if os.path.isdir(self.tmp):
            shutil.rmtree(self.tmp)

    # TESTS

    def test_basic_init_ws_abs(self):
        ''' test edk2path with valid absolute path to workspace'''
        pathobj = Edk2Path(self.tmp, [])
        self.assertEqual(pathobj.WorkspacePath, self.tmp)

    def test_basic_init_ws_cwd(self):
        ''' test edk2path with a relative path to workspace'''
        relpath = "testrootfolder"
        fullpath = os.path.join(self.tmp, relpath)
        os.mkdir(fullpath)
        pathobj = Edk2Path(relpath, [])
        self.assertEqual(pathobj.WorkspacePath, fullpath)

    def test_nonexistant_ws(self):
        ''' test edk2path with invalid workspace'''

        invalid_ws = os.path.join(self.tmp, "invalidpath")
        with self.assertRaises(Exception):
            Edk2Path(invalid_ws, [])

    def test_nonexistant_abs(self):
        ''' test edk2path with valid ws but invalid pp'''
        pp = "doesnot_exist"
        pp_full_1 = os.path.join(self.tmp, pp)
        # absolute path
        with self.assertRaises(Exception):
            Edk2Path(self.tmp, [pp_full_1])
        # relative path
        with self.assertRaises(Exception):
            Edk2Path(self.tmp, [pp])
        # confirm optional parameter to remove invalid pp values
        pathobj = Edk2Path(self.tmp, [pp], error_on_invalid_pp=False)
        self.assertEqual(len(pathobj.PackagePathList), 0)

    def test_pp_inside_workspace(self):
        ''' test with packagespath pointing to folder nested inside workspace
        root/                   <-- current working directory
            folder_ws/           <-- workspace root
                folder_pp/       <-- packages path
                    pp packages here
                ws packages here
        '''
        ws_rel = "folder_ws"
        ws_abs = os.path.join(self.tmp, ws_rel)
        os.mkdir(ws_abs)
        folder_pp_rel = "pp1"
        folder_pp1_abs = os.path.join(ws_abs, folder_pp_rel)
        os.mkdir(folder_pp1_abs)

        # pp absolute
        pathobj = Edk2Path(ws_abs, [folder_pp1_abs])
        self.assertEqual(pathobj.WorkspacePath, ws_abs)
        self.assertEqual(pathobj.PackagePathList[0], folder_pp1_abs)

        # pp relative to workspace
        pathobj = Edk2Path(ws_abs, [folder_pp_rel])
        self.assertEqual(pathobj.WorkspacePath, ws_abs)
        self.assertEqual(pathobj.PackagePathList[0], folder_pp1_abs)

        # pp relative to cwd
        pathobj = Edk2Path(ws_abs, [os.path.join(ws_rel, folder_pp_rel)])
        self.assertEqual(pathobj.WorkspacePath, ws_abs)
        self.assertEqual(pathobj.PackagePathList[0], folder_pp1_abs)

    def test_pp_outside_workspace(self):
        ''' test with packagespath pointing to folder outside of workspace
        root/                   <-- current working directory
            folder_ws/           <-- workspace root
                ws packages here
            folder_pp/       <-- packages path
                pp packages here

        '''
        ws_rel = "folder_ws"
        ws_abs = os.path.join(self.tmp, ws_rel)
        os.mkdir(ws_abs)
        folder_pp_rel = "pp1"
        folder_pp1_abs = os.path.join(self.tmp, folder_pp_rel)
        os.mkdir(folder_pp1_abs)

        # pp absolute
        pathobj = Edk2Path(ws_abs, [folder_pp1_abs])
        self.assertEqual(pathobj.WorkspacePath, ws_abs)
        self.assertEqual(pathobj.PackagePathList[0], folder_pp1_abs)

        # pp relative to cwd
        pathobj = Edk2Path(ws_abs, [folder_pp_rel])
        self.assertEqual(pathobj.WorkspacePath, ws_abs)
        self.assertEqual(pathobj.PackagePathList[0], folder_pp1_abs)

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_basic_init_ws_abs_different_case(self):
        inputPath = self.tmp.capitalize()
        if self.tmp[0].isupper():
            inputPath = self.tmp[0].lower() + self.tmp[1:]

        pathobj = Edk2Path(inputPath, [])
        self.assertNotEqual(pathobj.WorkspacePath, self.tmp)

    def test_get_containing_package_inside_workspace(self):
        ''' test basic usage of GetContainingPackage with packages path nested
        inside the workspace

        File layout:

         root/                  <-- current working directory (self.tmp)
            folder_ws/           <-- workspace root
                folder_pp/       <-- packages path
                    PPTestPkg/   <-- A edk2 package
                        PPTestPkg.DEC
                        module1/
                            module1.INF
                        module2/
                            module2.INF
                            X64/
                                TestFile.c
                WSTestPkg/   <-- A edk2 package
                    WSTestPkg.dec
                    module1/
                        module1.inf
                    module2/
                        module2.inf
                        X64/
                            TestFile.c
        '''
        ws_rel = "folder_ws"
        ws_abs = os.path.join(self.tmp, ws_rel)
        os.mkdir(ws_abs)
        folder_pp_rel = "pp1"
        folder_pp1_abs = os.path.join(ws_abs, folder_pp_rel)
        os.mkdir(folder_pp1_abs)
        ws_p_name = "WSTestPkg"
        ws_pkg_abs = self._make_edk2_package_helper(ws_abs, ws_p_name)
        pp_p_name = "PPTestPkg"
        pp_pkg_abs = self._make_edk2_package_helper(folder_pp1_abs, pp_p_name, extension_case_lower=False)
        pathobj = Edk2Path(ws_abs, [folder_pp1_abs])

        # file in WSTestPkg root
        p = os.path.join(ws_pkg_abs, "testfile.c")
        self.assertEqual(pathobj.GetContainingPackage(p), ws_p_name)

        # directory in WSTestPkg
        p = os.path.join(ws_pkg_abs, "X64")
        self.assertEqual(pathobj.GetContainingPackage(p), ws_p_name)

        # WSTestPkg directory itself
        self.assertEqual(pathobj.GetContainingPackage(ws_pkg_abs), ws_p_name)

        # file in module in WSTestPkg
        p = os.path.join(ws_pkg_abs, "module1", "testfile.c")
        self.assertEqual(pathobj.GetContainingPackage(p), ws_p_name)

        # file in workspace root - no package- should return None
        p = os.path.join(ws_abs, "testfile.c")
        self.assertIsNone(pathobj.GetContainingPackage(p))

        # file outside of the workspace - invalid and should return None
        p = os.path.join(os.path.dirname(ws_abs), "testfile.c")
        self.assertIsNone(pathobj.GetContainingPackage(p))

        # file in PPTestPkg root
        p = os.path.join(pp_pkg_abs, "testfile.c")
        self.assertEqual(pathobj.GetContainingPackage(p), pp_p_name)

        # file in module in WSTestPkg
        p = os.path.join(pp_pkg_abs, "module1", "testfile.c")
        self.assertEqual(pathobj.GetContainingPackage(p), pp_p_name)

        # file in packages path root - no package- should return None
        p = os.path.join(folder_pp1_abs, "testfile.c")
        self.assertIsNone(pathobj.GetContainingPackage(p))

    def test_get_containing_package_outside_workspace(self):
        ''' test basic usage of GetContainingPackage with packages path
        outside the workspace

        File layout:

         root/                  <-- current working directory (self.tmp)
            folder_ws/           <-- workspace root
                WSTestPkg/   <-- A edk2 package
                    WSTestPkg.dec
                    module1/
                        module1.inf
                    module2/
                        module2.inf
                        X64/
                            TestFile.c
            folder_pp/       <-- packages path
                PPTestPkg/   <-- A edk2 package
                    PPTestPkg.DEC
                    module1/
                        module1.INF
                    module2/
                        module2.INF
                        X64/
                            TestFile.c
        '''
        ws_rel = "folder_ws"
        ws_abs = os.path.join(self.tmp, ws_rel)
        os.mkdir(ws_abs)
        folder_pp_rel = "pp1"
        folder_pp1_abs = os.path.join(self.tmp, folder_pp_rel)
        os.mkdir(folder_pp1_abs)
        ws_p_name = "WSTestPkg"
        ws_pkg_abs = self._make_edk2_package_helper(ws_abs, ws_p_name)
        pp_p_name = "PPTestPkg"
        pp_pkg_abs = self._make_edk2_package_helper(folder_pp1_abs, pp_p_name, extension_case_lower=False)
        pathobj = Edk2Path(ws_abs, [folder_pp1_abs])

        # file in WSTestPkg root
        p = os.path.join(ws_pkg_abs, "testfile.c")
        self.assertEqual(pathobj.GetContainingPackage(p), ws_p_name)

        # file in module in WSTestPkg
        p = os.path.join(ws_pkg_abs, "module1", "testfile.c")
        self.assertEqual(pathobj.GetContainingPackage(p), ws_p_name)

        # file in workspace root - no package- should return None
        p = os.path.join(ws_abs, "testfile.c")
        self.assertIsNone(pathobj.GetContainingPackage(p))

        # file outside of the workspace - invalid and should return None
        p = os.path.join(os.path.dirname(ws_abs), "testfile.c")
        self.assertIsNone(pathobj.GetContainingPackage(p))

        # file in PPTestPkg root
        p = os.path.join(pp_pkg_abs, "testfile.c")
        self.assertEqual(pathobj.GetContainingPackage(p), pp_p_name)

        # file in module in WSTestPkg
        p = os.path.join(pp_pkg_abs, "module1", "testfile.c")
        self.assertEqual(pathobj.GetContainingPackage(p), pp_p_name)

        # file in packages path root - no package- should return None
        p = os.path.join(folder_pp1_abs, "testfile.c")
        self.assertIsNone(pathobj.GetContainingPackage(p))

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_get_containing_package_ws_abs_different_case(self):
        ''' test basic usage of GetContainingPackage when the workspace path has different case for
        the drive letter then the incoming paths. This can happen on Windows
        if os.path.realpath is used.

        File layout:

         root/                  <-- current working directory (self.tmp)
            folder_ws/           <-- workspace root
                folder_pp/       <-- packages path
                    PPTestPkg/   <-- A edk2 package
                        PPTestPkg.DEC
                        module1/
                            module1.INF
                        module2/
                            module2.INF
                            X64/
                                TestFile.c
                WSTestPkg/   <-- A edk2 package
                    WSTestPkg.dec
                    module1/
                        module1.inf
                    module2/
                        module2.inf
                        X64/
                            TestFile.c
        '''
        ws_rel = "folder_ws"
        ws_abs = os.path.join(self.tmp, ws_rel)
        wsi_abs = os.path.join(self.tmp, ws_rel.capitalize())  # invert the case of the first char of the ws folder name
        os.mkdir(ws_abs)
        folder_pp_rel = "pp1"
        folder_pp1_abs = os.path.join(ws_abs, folder_pp_rel)
        os.mkdir(folder_pp1_abs)
        ws_p_name = "WSTestPkg"
        ws_pkg_abs = self._make_edk2_package_helper(ws_abs, ws_p_name)
        pp_p_name = "PPTestPkg"
        pp_pkg_abs = self._make_edk2_package_helper(folder_pp1_abs, pp_p_name, extension_case_lower=False)
        pathobj = Edk2Path(wsi_abs, [folder_pp1_abs])

        # confirm inverted
        self.assertNotEqual(pathobj.WorkspacePath, ws_abs)
        self.assertEqual(os.path.normcase(pathobj.WorkspacePath), os.path.normcase(ws_abs))

        # file in WSTestPkg root
        p = os.path.join(ws_pkg_abs, "testfile.c")
        self.assertEqual(pathobj.GetContainingPackage(p), ws_p_name)

        # file in module in WSTestPkg
        p = os.path.join(ws_pkg_abs, "module1", "testfile.c")
        self.assertEqual(pathobj.GetContainingPackage(p), ws_p_name)

        # file in workspace root - no package- should return None
        p = os.path.join(ws_abs, "testfile.c")
        self.assertIsNone(pathobj.GetContainingPackage(p), ws_rel)

        # file outside of the workspace - invalid and should return None
        p = os.path.join(os.path.dirname(ws_abs), "testfile.c")
        self.assertIsNone(pathobj.GetContainingPackage(p))

        # file in PPTestPkg root
        p = os.path.join(pp_pkg_abs, "testfile.c")
        self.assertEqual(pathobj.GetContainingPackage(p), pp_p_name)

        # file in module in WSTestPkg
        p = os.path.join(pp_pkg_abs, "module1", "testfile.c")
        self.assertEqual(pathobj.GetContainingPackage(p), pp_p_name)

        # file in packages path root - no package- should return None
        p = os.path.join(folder_pp1_abs, "testfile.c")
        self.assertIsNone(pathobj.GetContainingPackage(p), folder_pp_rel)

    def test_get_containing_modules_with_relative_path(self):
        """Test that a relative path raises an exception.

        Note: GetContainingModules() only accepts absolute paths.

        File layout:
            root/                   <-- Current working directory (self.tmp)
                folder_ws           <-- Workspace directory
                    pp1             <-- Package Path 1
                        PPTestPkg   <-- An edk2 package
                            PPTestPkg.DEC
                            module1
                                module1.INF
                            module2
                                module2.INF
                                X64
                                    TestFile.c
                    WSTestPkg       <-- An edk2 package
                        WSTestPkg.dec
                        module1
                            module1.inf
                        module2
                            module2.inf
                            X64
                                TestFile.c
       """
        # Make the workspace directory: folder_ws/
        ws_rel = "folder_ws"
        ws_abs = os.path.join(self.tmp, ws_rel)
        os.mkdir(ws_abs)

        # Make Package Path 1 directory: folder_ws/pp1
        folder_pp_rel = "pp1"
        folder_pp1_abs = os.path.join(ws_abs, folder_pp_rel)
        os.mkdir(folder_pp1_abs)

        # Make WSTestPkg: folder_ws/WSTestPkg
        ws_p_name = "WSTestPkg"
        self._make_edk2_package_helper(ws_abs, ws_p_name)

        # Make PPTestPkg in Package Path 1: folder_ws/pp1/PPTestPkg
        pp_p_name = "PPTestPkg"
        self._make_edk2_package_helper(folder_pp1_abs, pp_p_name, extension_case_lower=False)

        pathobj = Edk2Path(ws_abs, [folder_pp1_abs])

        # Change the current working directory to the workspace
        os.chdir(ws_abs)

        # Pass a valid relative path to GetContainingModules()
        #   folder_ws/WSTestPkg/module2/module2.inf
        p = os.path.join("WSTestPkg", "module2", "module2.inf")
        self.assertRaises(Exception, pathobj.GetContainingModules, p)

        # Pass an invalid relative path to GetContainingModules()
        #   folder_ws/WSTestPkg/module2/module3.inf
        p = os.path.join("WSTestPkg", "module2", "module3.inf")
        self.assertRaises(Exception, pathobj.GetContainingModules, p)

        # Pass a valid non .inf relative path to GetContainingModules()
        #   folder_ws/WSTestPkg/module2/X64/TestFile.c
        p = os.path.join("WSTestPkg", "module2", "X64", "TestFile.c")
        self.assertRaises(Exception, pathobj.GetContainingModules, p)

    def test_get_containing_module_with_infs_in_other_temp_dirs(self):
        ''' test that GetContainingModule does not look outside the workspace
        root for modules. To do so, a temporary .inf file is placed in the
        user's temporary directory. Such a file could already exist and
        similarly impact test results. To ensure consistent test results, this
        test explicitly creates such a file.

        File layout:

        root/                         <-- User temporary directory
           SomeModule.inf             <-- .inf file in user temporary directory
           <temp_working_dir>/        <-- Current working directory (self.tmp)
               folder_ws              <-- Workspace root
                   WSTestPkg          <-- An edk2 package
                       WSTestPkg.dec
                       module1
                           module1.inf
                       module2
                           module2.inf
                           X64
                               TestFile.c
        '''
        # Make the workspace directory: <self.tmp>/folder_ws/
        ws_rel = "folder_ws"
        ws_abs = os.path.join(self.tmp, ws_rel)

        # Make WSTestPkg: <self.tmp>/folder_ws/WSTestPkg/
        ws_p_name = "WSTestPkg"
        self._make_edk2_package_helper(ws_abs, ws_p_name)

        # Place a .inf file in the temporary directory
        # <Temporary Directory>/SomeModule.inf
        other_inf = os.path.join(os.path.dirname(self.tmp), "SomeModule.inf")
        with open(other_inf, 'w'):
            pass

        try:
            pathobj = Edk2Path(ws_abs, [])

            # File outside of the workspace - invalid and should return None
            p = os.path.join(os.path.dirname(ws_abs), "testfile.c")
            relist = pathobj.GetContainingModules(p)
            self.assertEqual(len(relist), 0)
        finally:
            if os.path.isfile(other_inf):
                os.remove(other_inf)

    def test_get_containing_modules_path_format(self):
        """Test that get containing modules returns the exact expected path.

        Note: GetContainingModules() only accepts absolute paths.

        File layout:
            root/                   <-- Current working directory (self.tmp)
                folder_ws           <-- Workspace directory
                    pp1             <-- Package Path 1
                        PPTestPkg   <-- An edk2 package
                            PPTestPkg.DEC
                            module1
                                module1.INF
                            module2
                                module2.INF
                                X64
                                    TestFile.c
                    WSTestPkg       <-- An edk2 package
                        WSTestPkg.dec
                        module1
                            module1.inf
                        module2
                            module2.inf
                            X64
                                TestFile.c
        """
        # /folder_ws/
        ws_rel = "folder_ws"
        # /folder_ws/pp1/
        folder_pp_rel = "pp1"
        # /folder_ws/WSTestPkg/
        ws_p_name = "WSTestPkg"
        # /folder_ws/pp1/PPTestPkg
        pp_p_name = "PPTestPkg"

        # Create <temp_dir>/folder_ws/
        ws_abs = os.path.join(self.tmp, ws_rel)
        os.mkdir(ws_abs)

        # Create <temp_dir>/folder_ws/pp1/
        folder_pp1_abs = os.path.join(ws_abs, folder_pp_rel)
        os.mkdir(folder_pp1_abs)

        # Create <temp_dir>/folder_ws/WSTestPkg/
        ws_pkg_abs = self._make_edk2_package_helper(ws_abs, ws_p_name)

        # Create <temp_dir>/folder_ws/pp1/PPTestPkg
        pp_pkg_abs = self._make_edk2_package_helper(folder_pp1_abs, pp_p_name, extension_case_lower=False)
        pathobj = Edk2Path(ws_abs, [folder_pp1_abs])

        # Check the case and location of the absolute path
        p = os.path.join(pp_pkg_abs, "module1", "module1.INF")
        expected_module_inf = p
        actual_module_inf_list = pathobj.GetContainingModules(p)
        self.assertEqual(len(actual_module_inf_list), 1)
        self.assertEqual(expected_module_inf, actual_module_inf_list[0])

        # Check the case and location of the absolute path
        p = os.path.join(ws_pkg_abs, "module2", "X64", "TestFile.c")
        expected_module_inf = os.path.join(ws_pkg_abs, "module2", "module2.inf")
        actual_module_inf_list = pathobj.GetContainingModules(p)
        self.assertEqual(len(actual_module_inf_list), 1)
        self.assertEqual(expected_module_inf, actual_module_inf_list[0])

    def test_get_containing_module(self):
        ''' test basic usage of GetContainingModule with packages path nested
        inside the workspace

        File layout:

         root/                  <-- current working directory (self.tmp)
            folder_ws/           <-- workspace root
                folder_pp/       <-- packages path
                    PPTestPkg/   <-- A edk2 package
                        PPTestPkg.DEC
                        module1/
                            module1.INF
                        module2/
                            module2.INF
                            X64/
                                TestFile.c
                WSTestPkg/   <-- A edk2 package
                    WSTestPkg.dec
                    module1/
                        module1.inf
                    module2/
                        module2.inf
                        X64/
                            TestFile.c
        '''
        ws_rel = "folder_ws"
        ws_abs = os.path.join(self.tmp, ws_rel)
        os.mkdir(ws_abs)
        folder_pp_rel = "pp1"
        folder_pp1_abs = os.path.join(ws_abs, folder_pp_rel)
        os.mkdir(folder_pp1_abs)
        ws_p_name = "WSTestPkg"
        ws_pkg_abs = self._make_edk2_package_helper(ws_abs, ws_p_name)
        pp_p_name = "PPTestPkg"
        pp_pkg_abs = self._make_edk2_package_helper(folder_pp1_abs, pp_p_name, extension_case_lower=False)
        pathobj = Edk2Path(ws_abs, [folder_pp1_abs])

        # file in WSTestPkg root.  Module list is empty
        p = os.path.join(ws_pkg_abs, "testfile.c")
        relist = pathobj.GetContainingModules(p)
        self.assertEqual(len(relist), 0)

        # file in module in WSTestPkg/module1/module1.inf
        p = os.path.join(ws_pkg_abs, "module1", "testfile.c")
        relist = pathobj.GetContainingModules(p)
        self.assertEqual(len(relist), 1)
        self.assertEqual(
            Path(os.path.join(ws_pkg_abs, "module1", "module1.inf")),
            Path(relist[0]))

        # file in workspace root - no package- should return ws root
        p = os.path.join(ws_abs, "testfile.c")
        relist = pathobj.GetContainingModules(p)
        self.assertEqual(len(relist), 0)

        # file outside of the workspace - invalid and should return None
        p = os.path.join(os.path.dirname(ws_abs), "testfile.c")
        relist = pathobj.GetContainingModules(p)
        self.assertEqual(len(relist), 0)

        # file in module2 x64
        p = os.path.join(ws_pkg_abs, "module2", "X64", "testfile.c")
        relist = pathobj.GetContainingModules(p)
        self.assertEqual(len(relist), 1)
        self.assertEqual(
            Path(os.path.join(ws_pkg_abs, "module2", "module2.inf")),
            Path(relist[0]))

        # inf file in module2 x64
        p = os.path.join(ws_pkg_abs, "module2", "module2.inf")
        relist = pathobj.GetContainingModules(p)
        self.assertEqual(len(relist), 1)
        self.assertEqual(
            Path(os.path.join(ws_pkg_abs, "module2", "module2.inf")),
            Path(relist[0]))

        # file in PPTestPkg root
        p = os.path.join(pp_pkg_abs, "testfile.c")
        relist = pathobj.GetContainingModules(p)
        self.assertEqual(len(relist), 0)

        # file in module in PPTestPkg
        p = os.path.join(pp_pkg_abs, "module1", "testfile.c")
        relist = pathobj.GetContainingModules(p)
        self.assertEqual(len(relist), 1)
        self.assertEqual(
            Path(os.path.join(pp_pkg_abs, "module1", "module1.INF")),
            Path(relist[0]))

        # inf file in module in PPTestPkg
        p = os.path.join(pp_pkg_abs, "module1", "module1.INF")
        relist = pathobj.GetContainingModules(p)
        self.assertEqual(len(relist), 1)
        self.assertEqual(
            Path(os.path.join(pp_pkg_abs, "module1", "module1.INF")),
            Path(relist[0]))

        # file in packages path root - no module
        p = os.path.join(folder_pp1_abs, "testfile.c")
        relist = pathobj.GetContainingModules(p)
        self.assertEqual(len(relist), 0)

        # file doesn't exist and parent folder doesn't exist
        p = os.path.join(ws_pkg_abs, "ThisParentDirDoesNotExist", "ThisFileDoesNotExist.c")
        relist = pathobj.GetContainingModules(p)
        self.assertEqual(len(relist), 0)

        # file doesn't exist and parent folder doesn't exist and parent parent folder doesn't exist
        p = os.path.join(ws_pkg_abs, "DirDirDoesNotExist", "DirDoesNotExist", "FileDoesNotExist.c")
        relist = pathobj.GetContainingModules(p)
        self.assertEqual(len(relist), 0)

        # file doesn't exist and parent folder doesn't exist but parent parent is valid module
        # file in module in WSTestPkg/module1/module1.inf
        p = os.path.join(ws_pkg_abs, "module1", "ThisParentDirDoesNotExist", "testfile.c")
        relist = pathobj.GetContainingModules(p)
        self.assertEqual(len(relist), 1)
        self.assertEqual(
            Path(os.path.join(ws_pkg_abs, "module1", "module1.inf")),
            Path(relist[0]))

    def test_get_edk2_relative_path_from_absolute_path_posix(self):
        """Test that relative path returned is a POSIX path.

        File layout:
            root/                   <-- Current working directory (self.tmp)
                folder_ws           <-- Workspace directory
                    pp1             <-- Package Path 1
                        PPTestPkg   <-- An edk2 package
                            PPTestPkg.DEC
                            module1
                                module1.INF
                            module2
                                module2.INF
                                X64
                                    TestFile.c
                    WSTestPkg       <-- An edk2 package
                        WSTestPkg.dec
                        module1
                            module1.inf
                        module2
                            module2.inf
                            X64
                                TestFile.c
        """
        # /folder_ws/
        ws_rel = "folder_ws"
        # /folder_ws/pp1/
        folder_pp_rel = "pp1"
        # /folder_ws/WSTestPkg/
        ws_p_name = "WSTestPkg"
        # /folder_ws/pp1/PPTestPkg
        pp_p_name = "PPTestPkg"

        # Create <temp_dir>/folder_ws/
        ws_abs = os.path.join(self.tmp, ws_rel)
        os.mkdir(ws_abs)

        # Create <temp_dir>/folder_ws/pp1/
        folder_pp1_abs = os.path.join(ws_abs, folder_pp_rel)
        os.mkdir(folder_pp1_abs)

        # Create <temp_dir>/folder_ws/WSTestPkg/
        ws_pkg_abs = self._make_edk2_package_helper(ws_abs, ws_p_name)

        # Create <temp_dir>/folder_ws/pp1/PPTestPkg
        pp_pkg_abs = self._make_edk2_package_helper(folder_pp1_abs, pp_p_name, extension_case_lower=False)
        pathobj = Edk2Path(ws_abs, [folder_pp1_abs])

        p = os.path.join(pp_pkg_abs, "module1", "module1.INF")
        expected_rel_from_abs_path = PurePath(os.path.join(pp_p_name, "module1", "module1.INF")).as_posix()
        actual_rel_from_abs_path = pathobj.GetEdk2RelativePathFromAbsolutePath(p)
        self.assertEqual(expected_rel_from_abs_path, actual_rel_from_abs_path)

        p = os.path.join(ws_pkg_abs, "module_2", "X64", "TestFile.inf")
        expected_rel_from_abs_path = PurePath(os.path.join(ws_p_name, "module_2", "X64", "TestFile.inf")).as_posix()
        actual_rel_from_abs_path = pathobj.GetEdk2RelativePathFromAbsolutePath(p)
        self.assertEqual(expected_rel_from_abs_path, actual_rel_from_abs_path)

    def test_get_edk2_relative_path_from_absolute_path(self):
        ''' test basic usage of GetEdk2RelativePathFromAbsolutePath with packages path nested
        inside the workspace

        File layout:

         root/                  <-- current working directory (self.tmp)
            folder_ws/           <-- workspace root
                folder_pp/       <-- packages path
                    PPTestPkg/   <-- A edk2 package
                        PPTestPkg.DEC
                        module1/
                            module1.INF
                        module2/
                            module2.INF
                            X64/
                                TestFile.c
                WSTestPkg/   <-- A edk2 package
                    WSTestPkg.dec
                    module1/
                        module1.inf
                    module2/
                        module2.inf
                        X64/
                            TestFile.c
        '''
        ws_rel = "folder_ws"
        ws_abs = os.path.join(self.tmp, ws_rel)
        os.mkdir(ws_abs)
        folder_pp_rel = "pp1"
        folder_pp1_abs = os.path.join(ws_abs, folder_pp_rel)
        os.mkdir(folder_pp1_abs)
        ws_p_name = "WSTestPkg"
        ws_pkg_abs = self._make_edk2_package_helper(ws_abs, ws_p_name)
        pp_p_name = "PPTestPkg"
        pp_pkg_abs = self._make_edk2_package_helper(folder_pp1_abs, pp_p_name, extension_case_lower=False)
        pathobj = Edk2Path(ws_abs, [folder_pp1_abs])

        # file in packages path
        p = os.path.join(pp_pkg_abs, "module1", "module1.INF")
        self.assertEqual(pathobj.GetEdk2RelativePathFromAbsolutePath(p), f"{pp_p_name}/module1/module1.INF")

        # file in workspace
        p = os.path.join(ws_pkg_abs, "module2", "X64", "TestFile.c")
        self.assertEqual(pathobj.GetEdk2RelativePathFromAbsolutePath(p), f"{ws_p_name}/module2/X64/TestFile.c")

        # Folder in packages path
        p = os.path.join(pp_pkg_abs, "module2", "X64")
        self.assertEqual(pathobj.GetEdk2RelativePathFromAbsolutePath(p), f"{pp_p_name}/module2/X64")

        # Folder in workspace
        p = os.path.join(ws_pkg_abs, "module2", "X64")
        self.assertEqual(pathobj.GetEdk2RelativePathFromAbsolutePath(p), f"{ws_p_name}/module2/X64")

        # file not in workspace
        p = os.path.join(self.tmp, "module2", "X64", "TestFile.c")
        self.assertIsNone(pathobj.GetEdk2RelativePathFromAbsolutePath(p))

        # pass in bad parameter
        self.assertIsNone(pathobj.GetEdk2RelativePathFromAbsolutePath(None))

        # file is outside of code tree and not absolute path
        p = os.path.join("module2", "X64", "TestFile.c")
        self.assertIsNone(pathobj.GetEdk2RelativePathFromAbsolutePath(p))

        # file is cwd relative but not absolute path
        p = os.path.join(ws_rel, ws_p_name, "module2", "X64", "TestFile.c")
        self.assertIsNone(pathobj.GetEdk2RelativePathFromAbsolutePath(p))

        # ensure converted path keeps original capitalization
        p = os.path.join(ws_pkg_abs, "Module2", "x64", "TESTFILE.C")
        self.assertEqual(pathobj.GetEdk2RelativePathFromAbsolutePath(p), f"{ws_p_name}/Module2/x64/TESTFILE.C")

    def test_get_relative_path_when_packages_path_list_contains_substrings(self):
        ''' test usage of GetEdk2RelativePathFromAbsolutePath when members of PackagePathList contain
        substrings of themselves, for example "MU" and "MU_TIANO"
        File layout:
         root/                    <-- current working directory (self.tmp)
            folder_ws/            <-- workspace root
                folder_pp/        <-- packages path
                    PPTestPkg1/   <-- A edk2 package
                        PPTestPkg1.DEC
                        module1/
                            module1.INF
                            X64/
                                TestFile1.c
                folder_pp_suffix/ <-- packages path
                    PPTestPkg2/   <-- A edk2 package
                        PPTestPkg2.DEC
                        module2/
                            module2.INF
                            X64/
                                TestFile2.c
        '''
        ws_rel = "folder_ws"
        ws_abs = os.path.join(self.tmp, ws_rel)
        os.mkdir(ws_abs)
        folder_pp1_rel = "folder_pp"
        folder_pp1_abs = os.path.join(ws_abs, folder_pp1_rel)
        os.mkdir(folder_pp1_abs)
        folder_pp2_rel = "folder_pp_suffix"
        folder_pp2_abs = os.path.join(ws_abs, folder_pp2_rel)
        os.mkdir(folder_pp2_abs)
        ws_p_name = "PPTestPkg2"
        ws_pkg_abs = self._make_edk2_package_helper(folder_pp2_abs, ws_p_name)
        pathobj = Edk2Path(ws_abs, [folder_pp1_abs, folder_pp2_abs])

        # file in workspace
        p = os.path.join(ws_pkg_abs, "module2", "X64", "TestFile2.c")
        self.assertEqual(pathobj.GetEdk2RelativePathFromAbsolutePath(p), f"{ws_p_name}/module2/X64/TestFile2.c")

    def test_get_relative_path_when_package_path_inside_package(self):
        ''' test a package_path directory inside the subfolders of a package.
            Should not raise an exception.

        File layout:
        root/                            <-- Current working directory
            folder_ws/                   <-- Workspace root
                folder_pp1/              <-- Package Path 1
                    PPTestPkg1/          <-- A valid edk2 package
                        PPTestPkg1.dec
                        folder_pp2/      <-- Package Path 2
                        module1/
                            module1.inf
                        module2/
                            module2.inf
                            X64/
                                TestFile.c
        '''
        folder_ws_rel = "folder_ws"
        folder_ws_abs = os.path.join(self.tmp, folder_ws_rel)
        os.mkdir(folder_ws_abs)

        folder_pp1_rel = "folder_pp1"
        folder_pp1_abs = os.path.join(folder_ws_abs, folder_pp1_rel)
        os.mkdir(folder_pp1_abs)

        pp1_name = "PPTestPkg1"
        pp1_abs = self._make_edk2_package_helper(folder_pp1_abs, pp1_name)

        folder_pp2_rel = "folder_pp2"
        folder_pp2_abs = os.path.join(pp1_abs, folder_pp2_rel)
        os.mkdir(folder_pp2_abs)

        pathobj = Edk2Path(folder_ws_abs, [folder_pp1_abs, folder_pp2_abs])
        p = os.path.join(pp1_abs, "module2", "X64", "TestFile.c")
        self.assertEqual(pathobj.GetEdk2RelativePathFromAbsolutePath(p), f"{pp1_name}/module2/X64/TestFile.c")

    def test_get_relative_path_with_nested_packages(self):
        ''' test a two package paths that contain nested packages.
            Should raise an exception due to nested packages.

        File layout:
        root/                                   <-- Current working directory
            folder_ws/                          <-- Workspace root
                folder_pp1/                     <-- Package Path 1
                    PPTestPkg1/                 <-- A valid edk2 package
                        PPTestPkg1.dec
                        folder_pp2/             <-- Package Path 2
                            PPTestPkg2/         <-- A valid edk2 package
                                PPTestPkg2.dec
                                module1/
                                    module1.inf
                                module2/
                                    module2.inf
                                    X64/
                                        TestFile.c
                        module1/
                            module1.inf
                        module2/
                            module2.inf
                            X64/
                                TestFile.c
        '''
        folder_ws_rel = "folder_ws"
        folder_ws_abs = os.path.join(self.tmp, folder_ws_rel)
        os.mkdir(folder_ws_abs)

        folder_pp1_rel = "folder_pp1"
        folder_pp1_abs = os.path.join(folder_ws_abs, folder_pp1_rel)
        os.mkdir(folder_pp1_abs)

        pp1_name = "PPTestPkg1"
        pp1_abs = self._make_edk2_package_helper(folder_pp1_abs, pp1_name)

        folder_pp2_rel = "folder_pp2"
        folder_pp2_abs = os.path.join(pp1_abs, folder_pp2_rel)
        os.mkdir(folder_pp2_abs)

        pp2_name = "PPTestPkg2"
        self._make_edk2_package_helper(folder_pp2_abs, pp2_name)

        # Nested packages should raise an exception by default
        self.assertRaises(Exception, Edk2Path, folder_ws_abs, [folder_pp1_abs])
        self.assertRaises(Exception, Edk2Path, folder_ws_abs, [folder_pp1_abs, folder_pp2_abs])

        # Nested packages should no longer raise an exception
        # Note: These tests can be removed when support for
        #       PYTOOL_TEMPORARILY_IGNORE_NESTED_EDK_PACKAGES is removed.
        os.environ["PYTOOL_TEMPORARILY_IGNORE_NESTED_EDK_PACKAGES"] = "true"
        Edk2Path(folder_ws_abs, [folder_pp1_abs])
        Edk2Path(folder_ws_abs, [folder_pp1_abs, folder_pp2_abs])

        # Remove the environment variable now that the test above is complete
        os.environ.pop("PYTOOL_TEMPORARILY_IGNORE_NESTED_EDK_PACKAGES")

        # Nested packages should no longer raise an exception if explicitly
        # marked as known-bad.
        os.environ["PYTOOL_IGNORE_KNOWN_BAD_NESTED_PACKAGES"] = "SomeOtherPkg,PPTestPkg1"
        Edk2Path(folder_ws_abs, [folder_pp1_abs])
        Edk2Path(folder_ws_abs, [folder_pp1_abs, folder_pp2_abs])

        os.environ["PYTOOL_IGNORE_KNOWN_BAD_NESTED_PACKAGES"] = "SomeOtherPkg,PPTestPkg2"
        Edk2Path(folder_ws_abs, [folder_pp1_abs])
        Edk2Path(folder_ws_abs, [folder_pp1_abs, folder_pp2_abs])

        os.environ["PYTOOL_IGNORE_KNOWN_BAD_NESTED_PACKAGES"] = "SomeOtherPkg,SomeOtherPkg2"
        self.assertRaises(Exception, Edk2Path, folder_ws_abs, [folder_pp1_abs])
        self.assertRaises(Exception, Edk2Path, folder_ws_abs, [folder_pp1_abs, folder_pp2_abs])

        # Remove the environment variable now that the test above is complete
        os.environ.pop("PYTOOL_IGNORE_KNOWN_BAD_NESTED_PACKAGES")

    def test_get_relative_path_when_folder_is_next_to_package(self):
        ''' test usage of GetEdk2RelativePathFromAbsolutePath when a folder containing a package is in the same
        directory as a different package. This test ensures the correct value is returned regardless the order of
        the package paths.
        file layout:
        root/                      <-- Current working directory
            folder_ws/             <-- Workspace Root
                folder_pp1/        <-- A Package Path
                    PPTestPkg1     <-- A Package
                    folder_pp2/    <-- A Package Path
                        PPTestPkg2 <-- A Package
        '''
        folder_ws_rel = "folder_ws"
        folder_ws_abs = os.path.join(self.tmp, folder_ws_rel)
        os.mkdir(folder_ws_abs)

        folder_pp1_rel = "folder_pp1"
        folder_pp1_abs = os.path.join(folder_ws_abs, folder_pp1_rel)
        os.mkdir(folder_pp1_abs)

        pp1_name = "PPTestPkg1"
        pp1_abs = self._make_edk2_package_helper(folder_pp1_abs, pp1_name)
        p1 = os.path.join(pp1_abs, "module2", "X64", "TestFile2.c")

        folder_pp2_rel = "folder_pp2"
        folder_pp2_abs = os.path.join(folder_pp1_abs, folder_pp2_rel)
        os.mkdir(folder_pp2_abs)

        pp2_name = "PPTestPkg2"
        pp2_abs = self._make_edk2_package_helper(folder_pp2_abs, pp2_name)
        p2 = os.path.join(pp2_abs, "module2", "X64", "TestFile2.c")

        pathobj = Edk2Path(folder_ws_abs, [folder_pp1_abs, folder_pp2_abs])
        self.assertEqual(pathobj.GetEdk2RelativePathFromAbsolutePath(p1), f'{pp1_name}/module2/X64/TestFile2.c')
        self.assertEqual(pathobj.GetEdk2RelativePathFromAbsolutePath(p2), f'{pp2_name}/module2/X64/TestFile2.c')

        pathobj = Edk2Path(folder_ws_abs, [folder_pp2_abs, folder_pp1_abs])
        self.assertEqual(pathobj.GetEdk2RelativePathFromAbsolutePath(p1), f'{pp1_name}/module2/X64/TestFile2.c')
        self.assertEqual(pathobj.GetEdk2RelativePathFromAbsolutePath(p2), f'{pp2_name}/module2/X64/TestFile2.c')

    def test_get_relative_path_when_path_does_not_exist(self):
        ''' test basic usage of GetEdk2RelativePathFromAbsolutePath with packages path nested
        inside the workspace, but the absolute path is not a real path

        File layout:

         root/                  <-- current working directory (self.tmp)
            folder_ws/           <-- workspace root
                folder_pp/       <-- packages path
                    PPTestPkg/   <-- A edk2 package
                        PPTestPkg.DEC
                        module1/
                            module1.INF
                        module2/
                            module2.INF
                            X64/
                                TestFile.c
                WSTestPkg/   <-- A edk2 package
                    WSTestPkg.dec
                    module1/
                        module1.inf
                    module2/
                        module2.inf
                        X64/
                            TestFile.c
        '''
        ws_rel = "folder_ws"
        ws_abs = os.path.join(self.tmp, ws_rel)
        os.mkdir(ws_abs)
        folder_pp_rel = "pp1"
        folder_pp1_abs = os.path.join(ws_abs, folder_pp_rel)
        os.mkdir(folder_pp1_abs)
        ws_p_name = "WSTestPkg"
        ws_pkg_abs = self._make_edk2_package_helper(ws_abs, ws_p_name)
        pp_p_name = "PPTestPkg"
        pp_pkg_abs = self._make_edk2_package_helper(folder_pp1_abs, pp_p_name, extension_case_lower=False)
        pathobj = Edk2Path(ws_abs, [folder_pp1_abs])

        # not existant file in packages path
        p = os.path.join(pp_pkg_abs, "module1", "FAKE.INF")
        self.assertEqual(pathobj.GetEdk2RelativePathFromAbsolutePath(p), f"{pp_p_name}/module1/FAKE.INF")

        # non existant file in non existant directory in packages path
        p = os.path.join(ws_pkg_abs, "module2", "X64", "FAKE", "FAKE.c")
        self.assertEqual(pathobj.GetEdk2RelativePathFromAbsolutePath(p), f"{ws_p_name}/module2/X64/FAKE/FAKE.c")

        # not existant file in workspace
        p = os.path.join(ws_pkg_abs, "module2", "X64", "FAKE.c")
        self.assertEqual(pathobj.GetEdk2RelativePathFromAbsolutePath(p), f"{ws_p_name}/module2/X64/FAKE.c")

        # not existant file not in workspace
        p = os.path.join(self.tmp, "module2", "X64", "FAKE.c")
        self.assertIsNone(pathobj.GetEdk2RelativePathFromAbsolutePath(p))

        # not existant file is outside of code tree and not absolute path
        p = os.path.join("module2", "X64", "FAKE.c")
        self.assertIsNone(pathobj.GetEdk2RelativePathFromAbsolutePath(p))

        # not existant file is cwd relative but not absolute path
        p = os.path.join(ws_rel, ws_p_name, "module2", "X64", "FAKE.c")
        self.assertIsNone(pathobj.GetEdk2RelativePathFromAbsolutePath(p))

    def test_get_relative_path_when_package_is_not_directly_inside_packages_path(self):
        ''' test basic usage of GetEdk2RelativePathFromAbsolutePath when the
        package is not a direct subfolder of a packagespath, but atleast one
        folder away.

        File layout:

         root/                  <-- current working directory (self.tmp)
            folder_ws/           <-- workspace root
                folder_pp/       <-- packages path
                    folder_extra/
                        PPTestPkg/   <-- A edk2 package
                            PPTestPkg.DEC
        '''
        ws_rel = "folder_ws"
        ws_abs = os.path.join(self.tmp, ws_rel)
        os.mkdir(ws_abs)

        folder_pp_rel = "folder_pp"
        folder_pp_abs = os.path.join(ws_abs, folder_pp_rel)
        os.mkdir(folder_pp_abs)

        folder_extra_rel = "folder_extra"
        folder_extra_abs = os.path.join(folder_pp_abs, folder_extra_rel)
        os.mkdir(folder_extra_abs)

        ws_p_name = "PPTestPkg"
        ws_pkg_abs = self._make_edk2_package_helper(folder_extra_abs, ws_p_name)
        pathobj = Edk2Path(ws_abs, [folder_pp_abs])

        p = os.path.join(ws_pkg_abs, "PPTestPkg.dec")
        self.assertEqual(pathobj.GetEdk2RelativePathFromAbsolutePath(p),
                         f"{folder_extra_rel}/{ws_p_name}/{ws_p_name}.dec")

    def test_get_absolute_path_on_this_system_from_edk2_relative_path(self):
        ''' test basic usage of GetAbsolutePathOnThisSystemFromEdk2RelativePath with packages path nested
        inside the workspace

        File layout:

         root/                  <-- current working directory (self.tmp)
            folder_ws/           <-- workspace root
                folder_pp/       <-- packages path
                    PPTestPkg/   <-- A edk2 package
                        PPTestPkg.DEC
                        module1/
                            module1.INF
                        module2/
                            module2.INF
                            X64/
                                TestFile.c
                WSTestPkg/   <-- A edk2 package
                    WSTestPkg.dec
                    module1/
                        module1.inf
                    module2/
                        module2.inf
                        X64/
                            TestFile.c
        '''
        ws_rel = "folder_ws"
        ws_abs = os.path.join(self.tmp, ws_rel)
        os.mkdir(ws_abs)
        folder_pp_rel = "pp1"
        folder_pp1_abs = os.path.join(ws_abs, folder_pp_rel)
        os.mkdir(folder_pp1_abs)
        ws_p_name = "WSTestPkg"
        ws_pkg_abs = self._make_edk2_package_helper(ws_abs, ws_p_name)
        pp_p_name = "PPTestPkg"
        pp_pkg_abs = self._make_edk2_package_helper(folder_pp1_abs, pp_p_name, extension_case_lower=False)
        pathobj = Edk2Path(ws_abs, [folder_pp1_abs])

        # file in packages path
        ep = os.path.join(pp_pkg_abs, "module1", "module1.INF")
        rp = f"{pp_p_name}/module1/module1.INF"
        self.assertEqual(pathobj.GetAbsolutePathOnThisSystemFromEdk2RelativePath(rp), ep)

        # file in workspace
        ep = os.path.join(ws_pkg_abs, "module2", "X64", "TestFile.c")
        rp = f"{ws_p_name}/module2/X64/TestFile.c"
        self.assertEqual(pathobj.GetAbsolutePathOnThisSystemFromEdk2RelativePath(rp), ep)

        # file not in workspace
        rp = "DoesNotExistPkg/module2/X64/TestFile.c"
        self.assertIsNone(pathobj.GetAbsolutePathOnThisSystemFromEdk2RelativePath(rp))

        # pass in bad parameter
        self.assertIsNone(pathobj.GetAbsolutePathOnThisSystemFromEdk2RelativePath(None))

    def test_get_absolute_path_then_relative_path_when_path_contains_repeated_packagepath_name(self):
        ''' Test the back and forth between GetAbsolutePath and GetRelativeFromAbsolute when the
        path structure has multiple instances of a package path
        File layout:
         root/                  <-- current working directory (self.tmp)
            PlatformClientPkg/           <-- workspace root
                CLIENT_PKG/       <-- packages path
                    ClientPkg/   <-- A edk2 package
                        ClientPkg.DEC
                        module1/
                            module1.INF
                        module2/
                            module2.INF
                            x64/
                               TestFile.c
        '''
        ws_rel = "PlatformClientPkg"
        ws_abs = os.path.join(self.tmp, ws_rel)
        os.mkdir(ws_abs)

        folder_pp1_rel = "CLIENT_PKG"
        folder_pp1_abs = os.path.join(ws_abs, folder_pp1_rel)
        os.mkdir(folder_pp1_abs)

        ws_pkg_name = "ClientPkg"
        ws_pkg_abs = self._make_edk2_package_helper(folder_pp1_abs, ws_pkg_name)

        pathobj = Edk2Path(ws_abs, [folder_pp1_abs])

        # Check getting absolute path from relative path
        abspath = pathobj.GetAbsolutePathOnThisSystemFromEdk2RelativePath(
            os.path.join(ws_pkg_name, ws_pkg_name + ".dec"))

        self.assertEqual(abspath, os.path.join(ws_pkg_abs, "ClientPkg.dec"))

        # check get relative path from abs path
        expectedpath = f"{ws_pkg_name}/ClientPkg.dec"

        self.assertEqual(pathobj.GetEdk2RelativePathFromAbsolutePath(abspath), expectedpath)



# @file path_utilities.py
# Code to help convert Edk2, absolute, and relative file paths
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Code to help convert Edk2, absolute, and relative file paths."""
import os
import logging
import fnmatch
import errno
from typing import Iterable
from pathlib import Path


class Edk2Path(object):
    """Represents edk2 file paths.

    Class that helps perform path operations within an EDK workspace.

    There are two OS environment variables that modify the behavior of this class with
    respect to nested package checking:
        PYTOOL_TEMPORARILY_IGNORE_NESTED_EDK_PACKAGES - converts errors about nested
            packages to warnings.
        PYTOOL_IGNORE_KNOWN_BAD_NESTED_PACKAGES - contains a comma-delimited list of
            known bad packages. If a package matching a known bad package from the
            list is encountered, an info-level message will be printed and the nested
            package check will be skipped.

    """

    def __init__(self, ws: os.PathLike, package_path_list: Iterable[os.PathLike],
                 error_on_invalid_pp: bool = True):
        """Constructor.

        Args:
            ws (os.PathLike): absolute path or cwd relative path of the workspace.
            package_path_list (Iterable[os.PathLike]): list of packages path.
                Entries can be Absolute path, workspace relative path, or CWD relative.
            error_on_invalid_pp (bool): default value is True. If packages path
                                        value is invalid raise exception.

        Raises:
            (NotADirectoryError): Invalid workspace or package path directory.
        """
        self.WorkspacePath = ws
        self.logger = logging.getLogger("Edk2Path")

        # Other code is dependent the following types, so keep it that way:
        #   - self.PackagePathList: List[str]
        #   - self.WorkspacePath: str

        self.PackagePathList = []
        self.WorkspacePath = ""

        workspace_candidate_path = Path(ws)

        if not workspace_candidate_path.is_absolute():
            workspace_candidate_path = Path(os.getcwd(), ws)

        if not workspace_candidate_path.is_dir():
            raise NotADirectoryError(
                errno.ENOENT,
                os.strerror(errno.ENOENT),
                workspace_candidate_path.resolve())

        self.WorkspacePath = str(workspace_candidate_path)

        candidate_package_path_list = []
        for a in package_path_list:
            if os.path.isabs(a):
                candidate_package_path_list.append(Path(a))
            else:
                wsr = Path(self.WorkspacePath, a)
                if wsr.is_dir():
                    candidate_package_path_list.append(wsr)
                else:
                    # assume current working dir relative.  Will catch invalid dir when checking whole list
                    candidate_package_path_list.append(Path(os.getcwd(), a))

        error = False
        for a in candidate_package_path_list[:]:
            if not a.is_dir():
                self.logger.log(logging.ERROR if error_on_invalid_pp else
                                logging.WARNING,
                                f"Invalid package path entry {a.resolve()}")
                candidate_package_path_list.remove(a)
                error = True

        self.PackagePathList = [str(p) for p in candidate_package_path_list]

        if error and error_on_invalid_pp:
            raise NotADirectoryError(errno.ENOENT, os.strerror(errno.ENOENT),
                                     a.resolve())

        #
        # Nested package check - ensure packages do not exist in a linear
        # path hierarchy.
        #
        # 1. Build a dictionary for each package path.
        #      - Key = Package path
        #      - Value = List of packages discovered in package path
        # 2. Enumerate all keys in dictionary checking if any package
        #    is relative (nested) to each other.
        # 3. Raise an Exception if two packages are found to be nested.
        #
        package_path_packages = {}
        for package_path in candidate_package_path_list:
            package_path_packages[package_path] = \
                [Path(p).parent for p in package_path.glob('**/*.dec')]

        # Note: The ability to ignore this function raising an exception on
        #       nested packages is temporary. Do not plan on this variable
        #       being available long-term and try to resolve the nested
        #       packages problem right away.
        #
        # Removal is tracked in the following GitHub issue:
        # https://github.com/tianocore/edk2-pytool-library/issues/200
        ignore_nested_packages = False
        if "PYTOOL_TEMPORARILY_IGNORE_NESTED_EDK_PACKAGES" in os.environ and \
            os.environ["PYTOOL_TEMPORARILY_IGNORE_NESTED_EDK_PACKAGES"].strip().lower() == \
                "true":
            ignore_nested_packages = True

        if "PYTOOL_IGNORE_KNOWN_BAD_NESTED_PACKAGES" in os.environ:
            pkgs_to_ignore = os.environ["PYTOOL_IGNORE_KNOWN_BAD_NESTED_PACKAGES"].split(",")
        else:
            pkgs_to_ignore = []

        for package_path, packages in package_path_packages.items():
            packages_to_check = []
            for package in packages:
                if any(x in str(package) for x in pkgs_to_ignore):
                    self.logger.log(
                        logging.INFO,
                        f"Ignoring nested package check for known-bad package "
                        f"[{str(package)}].")
                else:
                    packages_to_check.append(package)
            for i, package in enumerate(packages_to_check):
                for j in range(i + 1, len(packages_to_check)):
                    comp_package = packages_to_check[j]
                    if (package.is_relative_to(comp_package)
                            or comp_package.is_relative_to(package)):
                        if ignore_nested_packages:
                            self.logger.log(
                                logging.WARNING,
                                f"Nested packages not allowed. The packages "
                                f"[{str(package)}] and [{str(comp_package)}] are "
                                f"nested.")
                            self.logger.log(
                                logging.WARNING,
                                "Note 1: Nested packages are being ignored right now because the "
                                "\"PYTOOL_TEMPORARILY_IGNORE_NESTED_EDK_PACKAGES\" environment variable "
                                "is set. Do not depend on this variable long-term.")
                            self.logger.log(
                                logging.WARNING,
                                "Note 2: Some pytool features may not work as expected with nested packages.")
                        else:
                            raise Exception(
                                f"Nested packages not allowed. The packages "
                                f"[{str(package)}] and [{str(comp_package)}] are "
                                f"nested. Set the \"PYTOOL_TEMPORARILY_IGNORE_NESTED_EDK_PACKAGES\" "
                                f"environment variable to \"true\" as a temporary workaround "
                                f"until you fix the packages so they are no longer nested.")

    def GetEdk2RelativePathFromAbsolutePath(self, abspath):
        """Given an absolute path return a edk2 path relative to workspace or packagespath.

        Note: absolute path must be in the OS specific path form
        Note: the relative path will be in POSIX-like path form

        Args:
            abspath (os.PathLike): absolute path to a file or directory. Path must contain OS specific separator.

        Returns:
            (os.PathLike): POSIX-like relative path to workspace or packagespath
            (None): abspath is none
            (None): path is not valid
        """
        if abspath is None:
            return None

        relpath = None
        found = False

        # Check if the Absolute path starts with any of the package paths. If a match is found, build the relative
        # path based off that package.
        #
        # Sort the package paths from from longest to shortest. This handles the case where a package and a package
        # path are in the same directory. See the following path_utilities_test for a detailed explanation of the
        # scenario: test_get_relative_path_when_folder_is_next_to_package
        for packagepath in sorted((os.path.normcase(p) for p in self.PackagePathList), reverse=True):

            # If a match is found, use the original string to avoid change in case
            if os.path.normcase(abspath).startswith(packagepath):
                self.logger.debug("Successfully converted AbsPath to Edk2Relative Path using PackagePath")
                relpath = abspath[len(packagepath):]
                found = True
                break

        # If a match was not found, check if absolute path is based on the workspace root.
        if not found and os.path.normcase(abspath).startswith(os.path.normcase(self.WorkspacePath)):
            self.logger.debug("Successfully converted AbsPath to Edk2Relative Path using WorkspacePath")
            relpath = abspath[len(self.WorkspacePath):]
            found = True

        if found:
            relpath = relpath.replace(os.sep, "/").strip("/")
            self.logger.debug(f'[{abspath}] -> [{relpath}]')
            return relpath

        # Absolute path was not in reference to a package path or the workspace root.
        self.logger.error("Failed to convert AbsPath to Edk2Relative Path")
        self.logger.error(f'AbsolutePath: {abspath}')
        return None

    def GetAbsolutePathOnThisSystemFromEdk2RelativePath(self, relpath, log_errors=True):
        """Given a edk2 relative path return an absolute path to the file in this workspace.

        Args:
            relpath (os.PathLike): POSIX-like path
            log_errors (:obj:`bool`, optional): whether to log errors

        Returns:
            (os.PathLike): absolute path in the OS specific form
            (None): invalid relpath
            (None): Unable to get the absolute path
        """
        if relpath is None:
            return None
        relpath = relpath.replace("/", os.sep)
        abspath = os.path.join(self.WorkspacePath, relpath)
        if os.path.exists(abspath):
            return abspath

        for a in self.PackagePathList:
            abspath = os.path.join(a, relpath)
            if (os.path.exists(abspath)):
                return abspath
        if log_errors:
            self.logger.error("Failed to convert Edk2Relative Path to an Absolute Path on this system.")
            self.logger.error("Relative Path: %s" % relpath)

        return None

    def GetContainingPackage(self, InputPath: str) -> str:
        """Find the package that contains the given path.

        This isn't perfect but at least identifies the directory consistently.

        Note: The inputPath must be in the OS specific path form.

        Args:
            InputPath (str): absolute path to a file, directory, or module.
                             supports both windows and linux like paths.

        Returns:
            (str): name of the package that the module is in.
        """
        self.logger.debug("GetContainingPackage: %s" % InputPath)
        # Make a list that has the path case normalized for comparison.
        # Note: This only does anything on Windows
        package_paths = [os.path.normcase(x) for x in self.PackagePathList]
        workspace_path = os.path.normcase(self.WorkspacePath)

        # 1. Handle the case that InputPath is not in the workspace tree
        path_root = None
        if workspace_path not in os.path.normcase(InputPath):
            for p in package_paths:
                if p in os.path.normcase(InputPath):
                    path_root = p
                    break
            if not path_root:
                return None

        # 2. Determine if the path is under a package in the workspace

        # Start the search within the first available directory. If provided InputPath is a directory, start there,
        # else (if InputPath is a file) move to it's parent directory and start there.
        if os.path.isdir(InputPath):
            dirpath = str(InputPath)
        else:
            dirpath = os.path.dirname(InputPath)

        if not path_root:
            path_root = workspace_path

        while path_root != os.path.normcase(dirpath):
            if os.path.exists(dirpath):
                for f in os.listdir(dirpath):
                    if fnmatch.fnmatch(f.lower(), '*.dec'):
                        a = os.path.basename(dirpath)
                        return a

            dirpath = os.path.dirname(dirpath)

        return None

    def GetContainingModules(self, input_path: str) -> list[str]:
        """Find the list of modules (inf files) for a file path.

        Note: This function only accepts absolute paths. An exception will
              be raised if a non-absolute path is given.

        Note: If input_path does not exist in the filesystem, this function
              will try to return the likely containing module(s) but if the
              entire module has been deleted, this isn't possible.

        - If a .inf file is given, that file is returned.
        - Otherwise, the nearest set of .inf files (in the closest parent)
          will be returned in a list of file path strings.

        Args:
            input_path (str): Absolute path to a file, directory, or module.
                              Supports both Windows and Linux like paths.

        Returns:
            (list[str]): Absolute paths of .inf files that could be the
                         containing module.
        """
        input_path = Path(input_path)
        if not input_path.is_absolute():
            # Todo: Return a more specific exception type when
            # https://github.com/tianocore/edk2-pytool-library/issues/184 is
            # implemented.
            raise Exception("Module path must be absolute.")

        package_paths = [Path(os.path.normcase(x)) for x in self.PackagePathList]
        workspace_path = Path(os.path.normcase(self.WorkspacePath))
        all_root_paths = package_paths + [workspace_path]

        # For each root path, find the maximum allowed root in its hierarchy.
        maximum_root_paths = all_root_paths
        for root_path in maximum_root_paths:
            for other_root_path in maximum_root_paths[:]:
                if root_path == other_root_path:
                    continue
                if root_path.is_relative_to(other_root_path):
                    if len(root_path.parts) > len(other_root_path.parts):
                        maximum_root_paths.remove(root_path)
                    else:
                        maximum_root_paths.remove(other_root_path)

        # Verify the file path is within a valid workspace or package path
        # directory.
        for path in maximum_root_paths:
            if input_path.is_relative_to(path):
                break
        else:
            return []

        modules = []
        if input_path.suffix == '.inf':
            # Return the file path given since it is a module .inf file
            modules = [str(input_path)]

        if not modules:
            # Continue to ascend directories up to a maximum root path.
            #
            # This handles cases like:
            #   ModuleDir/      |   ModuleDir/      | ...similarly nested files
            #     Module.inf    |     Module.inf    |
            #     x64/          |     Common/       |
            #       file.c      |       X64/        |
            #                   |         file.c    |
            #
            # The terminating condition of the loop is when a maximum root
            # path has been reached.
            #
            # A maximum root path represents the maximum allowed ascension
            # point in the input_path directory hierarchy as sub-roots like
            # a package path pointing under a workspace path are already
            # accounted for during maximum root path filtering.
            #
            # Given a root path is either the workspace or a package path,
            # neither of which are a module directory, once that point is
            # reached, all possible module candidates are exhausted.
            current_dir = input_path.parent
            while current_dir not in maximum_root_paths:
                if current_dir.is_dir():
                    current_dir_inf_files = \
                        [str(f) for f in current_dir.iterdir() if
                         f.is_file() and f.suffix.lower() == '.inf']
                    if current_dir_inf_files:
                        # A .inf file(s) exist in this directory.
                        #
                        # Since this is the closest parent that can be considered
                        # a module, return the .inf files as module candidates.
                        modules.extend(current_dir_inf_files)
                        break

                current_dir = current_dir.parent

        return modules



# @file dsc_test.py
# Tests for the data model for the EDK II DSC
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import unittest
from edk2toollib.uefi.edk2.build_objects.dsc import dsc
from edk2toollib.uefi.edk2.build_objects.dsc import library_class
from edk2toollib.uefi.edk2.build_objects.dsc import component
from edk2toollib.uefi.edk2.build_objects.dsc import definition
from edk2toollib.uefi.edk2.build_objects.dsc import dsc_buildoption_section_type
from edk2toollib.uefi.edk2.build_objects.dsc import dsc_pcd_section_type
from edk2toollib.uefi.edk2.build_objects.dsc import dsc_section_type


class TestDscObject(unittest.TestCase):

    def test_null_creation(self):
        d = dsc()
        self.assertNotEqual(d, None)

    def test_dsc_multple_defines(self):
        # When we add an object, it should overwrite the previous one
        d = TestDscObject.create_dsc_object()
        d.defines.add(definition("PLATFORM_NAME", "TEST2"))
        for define in d.defines:
            if define.name == "PLATFORM_NAME":  # check to make sure it matches
                self.assertEqual(define.value, "TEST2")

    def test_dsc_multple_library_classes(self):
        d = dsc()
        # When we add an object, it should overwrite the previous one
        common_section = dsc_section_type()
        d.library_classes[common_section].add(library_class("TEST", "BOB.inf"))
        self.assertEqual(len(d.library_classes[common_section]), 1)
        # we should override the previous one
        d.library_classes[common_section].add(library_class("TEST", "BOB2.inf"))
        self.assertEqual(len(d.library_classes[common_section]), 1)
        for lib in d.library_classes[common_section]:
            self.assertEqual(lib.inf, "BOB2.inf")  # make sure we overrode it
        self.assertEqual(len(d.library_classes[common_section]), 1)

        # make sure we can add a library to a different section and that
        IA32_section = dsc_section_type(arch="IA32")
        self.assertEqual(len(d.library_classes[IA32_section]), 0)
        d.library_classes[IA32_section].add(library_class("NULL", "BOB1.inf"))
        self.assertEqual(len(d.library_classes[IA32_section]), 1)
        d.library_classes[IA32_section].add(library_class("NULL", "BOB2.inf"))
        self.assertEqual(len(d.library_classes[IA32_section]), 2)

    def test_get_library_classes(self):
        ''' This serves more as an example of how to walk the DSC to get a library class for a componenet '''
        pass

    def test_put_in_bad_things(self):
        d = dsc()
        # make sure we can't add stuff to d.defines
        with self.assertRaises(ValueError):
            d.defines.add(library_class("NULL", "TEST.inf"))
        # make sure we can't add stuff to skus
        with self.assertRaises(ValueError):
            d.skus.add(library_class("TEST", "TEST.inf"))
        # make sure we can't add stuff to skus
        with self.assertRaises(ValueError):
            d.default_stores.add(component("TEST", "TEST.inf"))

        common_section = dsc_section_type()
        build_opt_section = dsc_buildoption_section_type()
        pcd_section = dsc_pcd_section_type("FEATUREFLAG")

        # now to check the build options
        d.build_options[build_opt_section] = set()
        with self.assertRaises(ValueError):
            d.build_options[pcd_section] = set()
        with self.assertRaises(ValueError):
            d.build_options[common_section] = set()
        with self.assertRaises(ValueError):
            d.build_options[build_opt_section].add(library_class("TEST", "TEST.inf"))
        # NEXTVER: once the adding logic is implemented, this will be need to redone
        with self.assertRaises(ValueError):
            d.build_options[build_opt_section] = set()

        # now to check the pcds
        d.pcds[pcd_section] = set()
        with self.assertRaises(ValueError):
            d.pcds[build_opt_section] = set()
        with self.assertRaises(ValueError):
            d.pcds[pcd_section].add(library_class("TEST", "TEST.inf"))
        # NEXTVER: once the adding logic is implemented, this will be need to redone
        with self.assertRaises(ValueError):
            d.pcds[pcd_section] = set()

        # now to check the library classes
        d.library_classes[common_section] = set()
        with self.assertRaises(ValueError):
            d.library_classes[build_opt_section] = set()
        with self.assertRaises(ValueError):
            d.library_classes[common_section].add(component("TEST.inf"))
        # NEXTVER: once the adding logic is implemented, this will be need to redone
        with self.assertRaises(ValueError):
            d.library_classes[common_section] = set()

        # now to check the components
        d.components[common_section] = set()
        with self.assertRaises(ValueError):
            d.components[build_opt_section] = set()
        with self.assertRaises(ValueError):
            d.components[common_section].add(library_class("TEST", "TEST.inf"))
        # NEXTVER: once the adding logic is implemented, this will be need to redone
        with self.assertRaises(ValueError):
            d.components[common_section] = set()

    @staticmethod
    def create_dsc_object():
        # Normally we would just read the dsc object
        d = dsc()
        # first add the defines
        d.defines.add(definition("PLATFORM_NAME", "TEST"))
        d.defines.add(definition("PLATFORM_GUID", "EB216561-961F-47EE-9EF9-CA426EF547C2"))
        d.defines.add(definition("OUTPUT_DIRECTORY", "Build/TEST"))
        d.defines.add(definition("SUPPORTED_ARCHITECTURES", "IA32 X64 AARCH64"))

        # Next add some library classes
        default_section = dsc_section_type()
        d.library_classes[default_section].add(library_class("NULL", "BOB.inf"))

        # Next add a component
        return d



# @file gen_api.py
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Python script used to automatically generate API Reference documentation.
Used in conjunction with mkdocs to generate static markdown files for each
file inside the edk2toollib package for ReadTheDocs hosting.
"""
import mkdocs_gen_files
import glob
import os


def main():
    """Entry into script that is executed."""
    files = glob.glob("**/*.py", recursive=True, root_dir="edk2toollib")

    excluded_files = ["__init__.py"]

    for file_path in files:
        edit_path = file_path
        # __init__ file excluded as they provide no API's that needs to be generated
        # tool files excluded as they have entire readmes on how to use the tool
        if file_path.split(os.sep)[-1] in excluded_files:
            continue

        # tests are excluded as no API reference is necessary
        if file_path.startswith("tests"):
            continue

        file_path = file_path.replace(".py", ".md")

        filename = f"api{os.sep}{file_path}"
        with mkdocs_gen_files.open(filename, "w") as f:
            ff = file_path.replace(os.sep, '.').replace('.md', '')
            ff = f"edk2toollib.{ff}"
            print(f"::: {ff}", file=f)
            print("    handler: python", file=f)
            print("    options:", file=f)
            print("        show_bases: False", file=f)
            print("        show_root_heading: True", file=f)
            print("        show_root_full_path: False", file=f)
            print("        show_signature_annotations: True", file=f)
            print("        separate_signature: True", file=f)
            print("        members_order: 'source'", file=f)
            print("        show_source: False", file=f)

        # Point the "Edit on Github" button in the docs to point at the source code
        edit_path = os.path.join('..', 'edk2toollib', edit_path)
        mkdocs_gen_files.set_edit_path(filename, edit_path)


main()



# @file base_parser_test.py
# Contains unit test routines for the base parser class.
#
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import unittest
from edk2toollib.uefi.edk2.parsers.base_parser import BaseParser
import tempfile
import os


class TestBaseParser(unittest.TestCase):

    def test_replace_boolean_constants(self):
        parser = BaseParser("")
        parser.SetInputVars({
            "true": "True",
            "false": "False",
            "b_true": True,
            "b_false": False
        })
        line = "$(true)"
        self.assertEqual(parser.ReplaceVariables(line), "TRUE")
        line = "$(false)"
        self.assertEqual(parser.ReplaceVariables(line), "FALSE")
        line = "$(b_true)"
        self.assertEqual(parser.ReplaceVariables(line), "TRUE")
        line = "$(b_false)"
        self.assertEqual(parser.ReplaceVariables(line), "FALSE")

    def test_replace_macro_using_dollarsign(self):
        parser = BaseParser("")
        parser.SetInputVars({
            "name": "sean"
        })
        line = "Hello $(name)!"
        self.assertEqual(parser.ReplaceVariables(line), "Hello sean!")

    def test_replace_macro_local_var_priority(self):
        parser = BaseParser("")
        parser.SetInputVars({
            "name": "sean"
        })
        parser.LocalVars["name"] = "fred"
        line = "Hello $(name)!"
        self.assertEqual(parser.ReplaceVariables(line), "Hello fred!")


class TestBaseParserConditionals(unittest.TestCase):

    def test_replace_macro_without_resolution(self):
        parser = BaseParser("")
        parser.SetInputVars({
            "name": "sean"
        })
        line = "!if $(Unknown_Token)!"
        self.assertEqual(parser.ReplaceVariables(line), "!if 0!")

    def test_replace_macro_ifdef_dollarsign(self):
        parser = BaseParser("")
        parser.SetInputVars({
            "name": "sean"
        })
        line = "!ifdef $(name)"
        self.assertEqual(parser.ReplaceVariables(line), "!ifdef sean")

        line = "!ifdef $(Invalid_Token)"
        self.assertEqual(parser.ReplaceVariables(line), "!ifdef 0")

        line = "!IFDEF $(name)"
        self.assertEqual(parser.ReplaceVariables(line), "!IFDEF sean")

    def test_replace_macro_ifndef_dollarsign(self):
        parser = BaseParser("")
        parser.SetInputVars({
            "name": "sean"
        })
        line = "!IfNDef $(name)"
        self.assertEqual(parser.ReplaceVariables(line), "!IfNDef sean")

        line = "!ifndef $(Invalid_Token)"
        self.assertEqual(parser.ReplaceVariables(line), "!ifndef 0")

        line = "!IFnDEF $(name)"
        self.assertEqual(parser.ReplaceVariables(line), "!IFnDEF sean")

    def test_replace_macro_ifdef(self):
        parser = BaseParser("")
        parser.SetInputVars({
            "name": "sean"
        })
        line = "!ifdef name"
        self.assertEqual(parser.ReplaceVariables(line), "!ifdef sean")

        line = "!ifdef Invalid_Token"
        self.assertEqual(parser.ReplaceVariables(line), "!ifdef 0")

        line = "!IFDEF name"
        self.assertEqual(parser.ReplaceVariables(line), "!IFDEF sean")

    def test_replace_macro_ifndef(self):
        parser = BaseParser("")
        parser.SetInputVars({
            "name": "sean"
        })
        line = "!IfNDef name"
        self.assertEqual(parser.ReplaceVariables(line), "!IfNDef sean")

        line = "!ifndef Invalid_Token"
        self.assertEqual(parser.ReplaceVariables(line), "!ifndef 0")

        line = "!IFnDEF name"
        self.assertEqual(parser.ReplaceVariables(line), "!IFnDEF sean")

    def test_replace_macro_elseif(self):
        parser = BaseParser("")
        parser.SetInputVars({
            "name": "matt"
        })
        line = "!elseif $(name)"
        self.assertEqual(parser.ReplaceVariables(line), "!elseif matt")

        line = "!ELSEIF $(Invalid_Token)"
        self.assertEqual(parser.ReplaceVariables(line), "!ELSEIF 0")

    def test_conditional_ifdef(self):
        parser = BaseParser("")

        # simple confirmation of expected behavior
        # don't need to test VariableReplacement
        # that is done in the tests replace methods
        self.assertTrue(parser.InActiveCode())
        parser.ProcessConditional("!ifdef 0")
        self.assertFalse(parser.InActiveCode())
        parser.PopConditional()

        self.assertTrue(parser.InActiveCode())
        parser.ProcessConditional("!ifdef blahblah")
        self.assertTrue(parser.InActiveCode())
        parser.PopConditional()

    def test_conditional_ifndef(self):
        parser = BaseParser("")

        # simple confirmation of expected behavior
        # don't need to test VariableReplacement
        # that is done in the tests replace methods
        self.assertTrue(parser.InActiveCode())
        parser.ProcessConditional("!ifndef 0")
        self.assertTrue(parser.InActiveCode())
        parser.PopConditional()

        self.assertTrue(parser.InActiveCode())
        parser.ProcessConditional("!ifndef blahblah")
        self.assertFalse(parser.InActiveCode())
        parser.PopConditional()

    def test_process_conditional_single_boolean(self):
        parser = BaseParser("")
        # check that nothing is on the stack
        self.assertEqual(len(parser.ConditionalStack), 0)
        # check that we're in active code
        self.assertTrue(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!IF TRUE"))
        # make sure we've added some to the stack - should we even be checking this?
        self.assertEqual(len(parser.ConditionalStack), 1)
        self.assertTrue(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!IF FALSE"))
        self.assertFalse(parser.InActiveCode())
        # make sure if pass in a true thing we aren't back in active
        self.assertTrue(parser.ProcessConditional("!IF TRUE"))
        self.assertFalse(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!EnDiF"))
        self.assertFalse(parser.InActiveCode())
        # pop off the false statement and make sure we're back to active code
        self.assertTrue(parser.ProcessConditional("!EnDiF"))
        self.assertTrue(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!EnDiF"))
        self.assertTrue(parser.InActiveCode())
        # check that nothing is on the stack
        self.assertEqual(len(parser.ConditionalStack), 0)
        # lower case false
        self.assertTrue(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!IF false"))
        self.assertFalse(parser.InActiveCode())

    def test_process_garbage_input(self):
        parser = BaseParser("")
        # make sure we fail the garbage input
        conditional_count = len(parser.ConditionalStack)
        self.assertFalse(parser.ProcessConditional("GARBAGE INPUT"))
        # make sure our count didn't change
        self.assertEqual(len(parser.ConditionalStack), conditional_count)

    def test_process_conditional_ands_ors(self):
        parser = BaseParser("")
        self.assertTrue(parser.ProcessConditional("!if TRUE == FALSE OR TRUE == TRUE"))
        # enable this once we have working and and or but for now just test it
        # self.assertTrue(parser.InActiveCode())
        # TODO: check for and and if once we've implemented this
        # check for nested things etc

    def test_process_extra_tokens(self):
        parser = BaseParser("")
        # make sure we can't do 5 tokens
        with self.assertRaises(RuntimeError):
            parser.ProcessConditional("!if 3 == 6 ==")

        # make sure we can't do three tokens
        with self.assertRaises(RuntimeError):
            parser.ProcessConditional("!if 3 ==")

    def test_process_conditional_hex_number(self):
        parser = BaseParser("")
        # check that a hex number doesn't equal itself
        self.assertTrue(parser.ProcessConditional("!IF 0x30 == 30"))
        self.assertFalse(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!endif"))
        # Check that two hex doesn't equal each other
        self.assertTrue(parser.ProcessConditional("!IF 0x20 == 0x30"))
        self.assertFalse(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!endif"))
        # check that hex equals decimal
        self.assertTrue(parser.ProcessConditional("!IF 0x20 == 32"))
        self.assertTrue(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!endif"))

    def test_process_conditional_greater_than(self):
        parser = BaseParser("")
        self.assertTrue(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!IF 30 > 50"))
        self.assertFalse(parser.InActiveCode())
        parser.ProcessConditional("!endif")
        self.assertTrue(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!IF 30 > 30"))
        self.assertFalse(parser.InActiveCode())
        parser.ProcessConditional("!endif")
        self.assertTrue(parser.ProcessConditional("!IF 50 > 30"))
        self.assertTrue(parser.InActiveCode())

    def test_process_conditional_less_than(self):
        parser = BaseParser("")
        self.assertTrue(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!IF 70 < 50"))
        self.assertFalse(parser.InActiveCode())
        parser.ProcessConditional("!endif")

        self.assertTrue(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!IF 70 < 70"))
        self.assertFalse(parser.InActiveCode())
        parser.ProcessConditional("!endif")

        self.assertTrue(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!IF 50 < 70"))
        self.assertTrue(parser.InActiveCode())

    def test_process_conditional_greater_than_equal(self):
        parser = BaseParser("")
        self.assertTrue(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!IF 30 >= 50"))
        self.assertFalse(parser.InActiveCode())
        parser.ProcessConditional("!endif")

        self.assertTrue(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!IF 30 >= 30"))
        self.assertTrue(parser.InActiveCode())
        parser.ProcessConditional("!endif")

        self.assertTrue(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!IF 50 >= 30"))
        self.assertTrue(parser.InActiveCode())

    def test_process_conditional_less_than_equal(self):
        parser = BaseParser("")
        self.assertTrue(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!IF 70 <= 50"))
        self.assertFalse(parser.InActiveCode())
        parser.ProcessConditional("!endif")

        self.assertTrue(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!IF 70 <= 70"))
        self.assertTrue(parser.InActiveCode())
        parser.ProcessConditional("!endif")

        self.assertTrue(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!IF 50 <= 70"))
        self.assertTrue(parser.InActiveCode())
        parser.ProcessConditional("!endif")

    def test_process_conditional_true_not_equals_false(self):
        parser = BaseParser("")
        # check != with true and false
        self.assertTrue(parser.ProcessConditional("!IF TRUE != FALSE"))
        self.assertTrue(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!EnDiF"))

    def test_process_conditional_not_equals_true_false(self):
        parser = BaseParser("")
        # check != with true and false
        self.assertTrue(parser.ProcessConditional("!IF RACECAR != RACECAR"))
        self.assertFalse(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!EnDiF"))
        self.assertTrue(parser.ProcessConditional("!IF false != FALSE"))
        self.assertFalse(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!EnDiF"))

    def test_process_conditional_false_equals_zero(self):
        parser = BaseParser("")
        self.assertTrue(parser.ProcessConditional("!IF FALSE == 0"))
        self.assertTrue(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!EnDiF"))

    def test_process_conditional_true_equals_one(self):
        parser = BaseParser("")
        # check != with true and false
        self.assertTrue(parser.ProcessConditional("!IF TRUE == 1"))
        self.assertTrue(parser.InActiveCode())
        self.assertTrue(parser.ProcessConditional("!EnDiF"))

    def test_process_conditional_true_cannot_be_greater_than(self):
        parser = BaseParser("")
        # check != with true and false
        with self.assertRaises(ValueError):
            parser.ProcessConditional("!IF TRUE >= 1")

    def test_process_conditional_true_cannot_be_greater_than_hex(self):
        parser = BaseParser("")
        # check != with true and false
        with self.assertRaises(ValueError):
            parser.ProcessConditional("!IF 0x7 >= TRUE")

    def test_process_conditional_non_numerical(self):
        parser = BaseParser("")
        # check non numerical values
        with self.assertRaises(ValueError):
            parser.ProcessConditional("!IF ROCKETSHIP > 50")
        with self.assertRaises(ValueError):
            parser.ProcessConditional("!if 50 < ROCKETSHIPS")

    def test_process_conditional_invalid_operators(self):
        parser = BaseParser("")
        # check weird operators
        with self.assertRaises(RuntimeError):
            self.assertTrue(parser.ProcessConditional("!IF 50 <> 50"))

    def test_process_bad_else(self):
        parser = BaseParser("")
        # check to make sure we can't do a malformed endif
        with self.assertRaises(RuntimeError):
            self.assertTrue(parser.ProcessConditional("!else test"))
        # try to pop the empty stack and invert it
        with self.assertRaises(IndexError):
            self.assertTrue(parser.ProcessConditional("!else"))

    def test_process_else(self):
        parser = BaseParser("")
        # check to make sure we can't do a malformed endif
        self.assertTrue(parser.ProcessConditional("!if TRUE"))
        self.assertTrue(parser.ProcessConditional("!else"))
        self.assertFalse(parser.InActiveCode())

    def test_process_bad_endif(self):
        parser = BaseParser("")
        # check to make sure we can't do a malformed endif
        with self.assertRaises(RuntimeError):
            self.assertTrue(parser.ProcessConditional("!endif test"))
        # try to pop the empty stack
        with self.assertRaises(IndexError):
            self.assertTrue(parser.ProcessConditional("!endif"))

    def test_process_conditional_variables(self):
        parser = BaseParser("")
        # check variables
        with self.assertRaises(RuntimeError):
            parser.ProcessConditional("!ifdef")
        with self.assertRaises(RuntimeError):
            parser.ProcessConditional("!ifndef")

    def test_process_conditional_reset(self):
        parser = BaseParser("")
        # test reset
        self.assertTrue(parser.ProcessConditional("!IF FALSE"))
        parser.ResetParserState()
        self.assertTrue(parser.InActiveCode())
        self.assertEqual(len(parser.ConditionalStack), 0)

    def test_process_in_conditional(self):
        parser = BaseParser("")
        parser.SetInputVars({"TOOL_CHAIN_TAG": "GCC5_TEST"})
        self.assertTrue(parser.ProcessConditional(
            '!if ("GCC49" in $(TOOL_CHAIN_TAG)) OR ("GCC5" in $(TOOL_CHAIN_TAG))'))
        self.assertTrue(parser.InActiveCode())
        parser.ResetParserState()
        parser.SetInputVars({"TOOL_CHAIN_TAG": "TESTGCC49"})
        self.assertTrue(parser.ProcessConditional(
            '!if ("GCC49" in $(TOOL_CHAIN_TAG)) OR ("GCC5" in $(TOOL_CHAIN_TAG))'))
        self.assertTrue(parser.InActiveCode())
        parser.ResetParserState()
        # Don't give it a tool chain tag that isn't in the things we're searching for
        parser.SetInputVars({"TOOL_CHAIN_TAG": "NOTFOUND"})
        self.assertTrue(parser.ProcessConditional(
            '!if ("GCC49" in $(TOOL_CHAIN_TAG)) OR ("GCC5" in $(TOOL_CHAIN_TAG))'))
        self.assertFalse(parser.InActiveCode())

    def test_process_or_operation_conditional(self):
        parser = BaseParser("")
        self.assertTrue(parser.EvaluateConditional('!IF TRUE OR FALSE'))
        self.assertTrue(parser.EvaluateConditional('!if FALSE OR TRUE'))
        self.assertTrue(parser.EvaluateConditional('!if FALSE || TRUE'))
        self.assertTrue(parser.EvaluateConditional('!if TRUE OR TRUE'))
        self.assertFalse(parser.EvaluateConditional('!if FALSE OR FALSE'))
        self.assertFalse(parser.EvaluateConditional('!if FALSE || FALSE'))

    def test_process_and_operation_conditional(self):
        parser = BaseParser("")
        self.assertFalse(parser.EvaluateConditional('!if TRUE AND FALSE'))
        self.assertFalse(parser.EvaluateConditional('!if FALSE AND TRUE'))
        self.assertTrue(parser.EvaluateConditional('!if TRUE AND TRUE'))
        self.assertTrue(parser.EvaluateConditional('!if TRUE && TRUE'))
        self.assertFalse(parser.EvaluateConditional('!if FALSE AND FALSE'))
        self.assertFalse(parser.EvaluateConditional('!if FALSE && FALSE'))

    def test_process_invalid_conditional(self):
        parser = BaseParser("")
        with self.assertRaises(RuntimeError):
            parser.EvaluateConditional('!if TRUE AND FALSE AND')
        with self.assertRaises(RuntimeError):
            parser.EvaluateConditional('TRUE AND FALSE AND')

    def test_emulator_conditional_or_double_in(self):
        parser = BaseParser("")
        parser.SetInputVars({"ARCH": "X64"})
        self.assertTrue(parser.ProcessConditional(
            '!if "IA32" in $(ARCH) || "X64" in $(ARCH)'))
        self.assertTrue(parser.InActiveCode())
        parser.ResetParserState()
        parser.SetInputVars({"ARCH": "IA32"})
        self.assertTrue(parser.ProcessConditional(
            '!if "IA32" in $(ARCH) || "X64" in $(ARCH)'))
        self.assertTrue(parser.InActiveCode())
        parser.ResetParserState()

    def test_emulator_conditional_not_in(self):
        parser = BaseParser("")
        parser.SetInputVars({"TOOL_CHAIN_TAG": "VS2019"})
        self.assertTrue(parser.ProcessConditional('!if "XCODE5" not in $(TOOL_CHAIN_TAG)'))
        self.assertTrue(parser.InActiveCode())
        parser.ResetParserState()

    def test_emulator_conditional_parens_order(self):
        ''' Makes sure the parenthesis affect the order of expressions '''
        parser = BaseParser("")
        self.assertFalse(parser.EvaluateConditional('!if TRUE OR FALSE AND FALSE'))
        self.assertTrue(parser.EvaluateConditional('!if TRUE OR (FALSE AND FALSE)'))
        parser.ResetParserState()

    def test_emulator_conditional_not_or(self):
        ''' Makes sure we can use the not with other operators '''
        parser = BaseParser("")
        self.assertTrue(parser.EvaluateConditional('!if FALSE NOT OR FALSE'))
        self.assertFalse(parser.EvaluateConditional('!if TRUE NOT OR FALSE'))
        self.assertFalse(parser.EvaluateConditional('!if FALSE NOT OR TRUE'))
        self.assertFalse(parser.EvaluateConditional('!if TRUE NOT OR TRUE'))

    def test_emulator_conditional_not_it_all(self):
        ''' Makes sure the parenthesis affect the order of expressions '''
        parser = BaseParser("")
        self.assertTrue(parser.EvaluateConditional('!if NOT FALSE OR FALSE'))
        self.assertFalse(parser.EvaluateConditional('!if NOT TRUE OR FALSE'))
        # check to make sure it works with parenthesis
        self.assertFalse(parser.EvaluateConditional('!if NOT(TRUE)'))
        self.assertTrue(parser.EvaluateConditional('!if NOT(FALSE)'))
        # make sure it works with the bang symbol
        self.assertFalse(parser.EvaluateConditional('!if !(TRUE)'))
        self.assertTrue(parser.EvaluateConditional('!if !(FALSE)'))

        parser.ResetParserState()


class TestBaseParserGuids(unittest.TestCase):

    def test_is_guid(self):
        guid1 = "= { 0xD3B36F2C, 0xD551, 0x11D4, {0x9A, 0x46, 0x0, 0x90, 0x27, 0x3F, 0xC1,0xD }}"
        parser = BaseParser("")
        self.assertTrue(parser.IsGuidString(guid1))
        guid2 = "= { 0xD3B36F2C, 0xD551, 0x11D4, {0x9A, 0x46, 0x0, 0x90, 0x27, 0x3F, 0xC1,0xD }"
        self.assertFalse(parser.IsGuidString(guid2))
        guid3 = "= { 0xD3B36F2C, 0xD551, 0x11D4, {0x9A, 0x0, 0x90, 0x27, 0x3F, 0xC1,0xD }}"
        self.assertFalse(parser.IsGuidString(guid3))
        # guid4 = "= { 0xD3B36F, 0xD551, 0x11D4, {0x9A, 0x46, 0x0, 0x90, 0x27, 0x3F, 0xC1,0xD }}"
        # TODO make sure we are checking length?
        # self.assertFalse(parser.IsGuidString(guid4))
        guid5 = " { 0xD3B36F2C, 0xD551, 0x11D4, {0x9A, 0x46, 0x0, 0x90, 0x27, 0x3F, 0xC1,0xD }}"
        self.assertFalse(parser.IsGuidString(guid5))

    def test_parse_guid(self):
        guid1 = "{ 0xD3B36F2C, 0xD551, 0x11D4, {0x9A, 0x46, 0x0, 0x90, 0x27, 0x3F, 0xC1,0xD }}"
        guid1_answer = "D3B36F2C-D551-11D4-9A46-0090273FC10D"
        parser = BaseParser("")
        guid1_result = parser.ParseGuid(guid1)
        self.assertEqual(guid1_answer, guid1_result)
        # try a bad guid and make sure it fails since it's missing an element
        guid2 = "{ 0xD3B36F2C, 0xD551, 0x11D4, { 0x9A, 0x46, 0x00, 0x90, 0x27, 0x3F, 0xC1 }}"
        with self.assertRaises(RuntimeError):
            parser.ParseGuid(guid2)

        # check one that's too long
        guid3 = "{ 0xD3B36FbadC, 0xD551, 0x11D4, { 0x9A, 0x46, 0x00, 0x90, 0x27, 0x3F, 0xC1, 0x4D }}"
        with self.assertRaises(RuntimeError):
            parser.ParseGuid(guid3)

        # check one that's too short
        guid4 = "{ 0x3, 0x1, 0x4, { 0xA, 0x6, 0x0, 0x9, 0x2, 0xF, 0x1, 0xD }}"
        guid4_answer = "00000003-0001-0004-0A06-0009020F010D"
        guid4_result = parser.ParseGuid(guid4)
        self.assertEqual(guid4_result, guid4_answer)


class TestBaseParserVariables(unittest.TestCase):

    def test_replace_input_variables(self):
        parser = BaseParser("")
        variables = {
            "FIFTY": 50,
            "TEST": "TEST",
            "LOWER NUMBER": 40,
            "HEX": "0x20",
            "BOOLEAN TRUE": "TRUE",
            "BOOLEAN FALSE": "FALSE",
        }
        parser.SetInputVars(variables)
        # check to make sure we don't modify if we don't have variables
        no_var = "this has no variables"
        no_var_result = parser.ReplaceVariables(no_var)
        self.assertEqual(no_var_result, no_var)
        # make sure we don't fail when we have unknown variables
        na_var = "unknown var $(UNKNOWN)"
        na_var_after = "unknown var $(UNKNOWN)"
        na_var_result = parser.ReplaceVariables(na_var)
        self.assertEqual(na_var_result, na_var_after)
        # make sure we're good for all the variables
        variable_str = "var $(%s)"
        for variable_key in variables:
            line = variable_str % variable_key
            result = parser.ReplaceVariables(line)
            val = "var " + str(variables[variable_key])
            self.assertEqual(result, val)

    def test_replace_local_variables(self):
        parser = BaseParser("")
        variables = {
            "FIFTY": 50,
            "TEST": "TEST",
            "LOWER NUMBER": 40,
            "HEX": "0x20",
            "BOOLEAN": "TRUE",
            "BOOLEAN FALSE": "FALSE",
        }
        parser.LocalVars = variables
        # check to make sure we don't modify if we don't have variables
        no_var = "this has no variables"
        no_var_result = parser.ReplaceVariables(no_var)
        self.assertEqual(no_var_result, no_var)
        # make sure we don't fail when we have unknown variables
        na_var = "unknown var $(UNKNOWN)"
        na_var_after = "unknown var $(UNKNOWN)"
        na_var_result = parser.ReplaceVariables(na_var)
        self.assertEqual(na_var_result, na_var_after)
        # make sure we're good for all the variables
        variable_str = "var $(%s)"
        for variable_key in variables:
            line = variable_str % variable_key
            result = parser.ReplaceVariables(line)
            val = "var " + str(variables[variable_key])
            self.assertEqual(result, val)


class TestBaseParserPathAndFile(unittest.TestCase):

    # because of how this works we use WriteLines, SetAbsPath, and SetPackagePath
    def test_find_path(self):
        # we're using write lines to make sure everything wo
        parser = BaseParser("")
        parser.Lines = ["hello"]
        package_paths = ["Common/Test", "SM_MAGIC"]
        root_path = tempfile.mkdtemp()
        index = 0
        # create the packages path folders
        for package in package_paths:
            pack_path = os.path.join(root_path, package)
            os.makedirs(pack_path)
            parser.WriteLinesToFile(os.path.join(pack_path, f"package_{index}.txt"))
            index += 1
        # setup the parser
        parser.SetBaseAbsPath(root_path)
        parser.SetPackagePaths(package_paths)

        # create the root and target files
        root_file = "root.txt"
        target_file = "target.txt"

        root_filepath = os.path.join(root_path, root_file)
        target_filedir = os.path.join(root_path, "BuildPkg")
        target_filepath = os.path.join(target_filedir, target_file)
        # create root file
        parser.WriteLinesToFile(root_filepath)
        # create target file
        os.makedirs(target_filedir)
        parser.WriteLinesToFile(target_filepath)
        parser.TargetFilePath = target_filepath
        # check if we can find the root
        root_found = parser.FindPath(root_file)
        self.assertEqual(root_found, root_filepath)
        # check we can find the target using the target path
        target_found = parser.FindPath(target_file)
        self.assertEqual(target_found, target_filepath)

        # check package relative packages
        for index in range(len(package_paths)):
            file_name = f"package_{index}.txt"
            pp_found = parser.FindPath(file_name)
            self.assertTrue(os.path.exists(pp_found))

        # invalid files
        invalid_filename = "YOU_WONT_FIND_ME.txt"
        invalid_file = os.path.join(root_path, invalid_filename)
        invalid_result = parser.FindPath(invalid_file)
        invalid_result2 = parser.FindPath(invalid_filename)
        self.assertEqual(None, invalid_result)
        self.assertEqual(None, invalid_result2)
        invalid_result3 = parser.FindPath(None)
        self.assertEqual(None, invalid_result3)

    # make sure we can write out to a file

    def test_write_lines(self):
        parser = BaseParser("")
        parser.Lines = ["hello"]
        root_path = tempfile.mkdtemp()
        file_path = os.path.join(root_path, "lines.txt")
        parser.WriteLinesToFile(file_path)
        self.assertTrue(os.path.exists(file_path))
        # TODO check to make sure that the file matches what we expect



# @file guid_parser_test.py
# Contains unit test routines for the guid parser class.
#
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import unittest
from edk2toollib.uefi.edk2.parsers.guid_parser import GuidParser


class TestGuidParser(unittest.TestCase):

    def test_valid_input_guid(self):
        SAMPLE_DATA_C_FORMAT_GUID = "{0x66341ae8, 0x668f, 0x4192, {0xb4, 0x4d, 0x5f, 0x87, 0xb8, 0x68, 0xf0, 0x41}}"  # noqa: E501
        SAMPLE_DATA_REG_FORMAT_GUID = "66341ae8-668f-4192-b44d-5f87b868f041"
        self.assertEqual(GuidParser.reg_guid_from_c_format(SAMPLE_DATA_C_FORMAT_GUID), SAMPLE_DATA_REG_FORMAT_GUID)
        self.assertEqual(GuidParser.c_guid_from_reg_format(SAMPLE_DATA_REG_FORMAT_GUID), SAMPLE_DATA_C_FORMAT_GUID)

        uuid_from_c = GuidParser.uuid_from_guidstring(SAMPLE_DATA_C_FORMAT_GUID)
        uuid_from_reg = GuidParser.uuid_from_guidstring(SAMPLE_DATA_REG_FORMAT_GUID)

        self.assertEqual(GuidParser.reg_guid_str_from_uuid(uuid_from_c),
                         GuidParser.reg_guid_str_from_uuid(uuid_from_reg))

        self.assertEqual(GuidParser.c_guid_str_from_uuid(uuid_from_c),
                         GuidParser.c_guid_str_from_uuid(uuid_from_reg))

    def test_invalid_reg_format_to_uuid(self):
        SAMPLE_DATA_REG_FORMAT_GUID = "66341ae8-668f4192b44d-0087b868f041"
        u = GuidParser.uuid_from_guidstring(SAMPLE_DATA_REG_FORMAT_GUID)
        self.assertIsNone(u)

    def test_invalid_reg_format_to_c_format(self):
        SAMPLE_DATA_REG_FORMAT_GUID = "66341ae8-668f4192b44d-0087b868f041"
        u = GuidParser.c_guid_from_reg_format(SAMPLE_DATA_REG_FORMAT_GUID)
        self.assertEqual('', u)

    def test_invalid_c_format_to_uuid(self):
        SAMPLE_DATA_C_FORMAT_GUID = "{0x66341ae8, 0x668f 0x4192 {0xb4, 0x4d, 0x5f, 0x87, 0xb8, 0x68, 0xf0, 0x41}}"
        u = GuidParser.uuid_from_guidstring(SAMPLE_DATA_C_FORMAT_GUID)
        self.assertIsNone(u)

    def test_invalid_c_format_to_reg(self):
        SAMPLE_DATA_C_FORMAT_GUID = "{0x66341ae8, 0x668f4192, 0x1234, {0xb4, 0x4d, 0x5f34, 0x87, 0xb8, 0x68, 0xf0, 0x41}}"  # noqa: E501
        u = GuidParser.reg_guid_from_c_format(SAMPLE_DATA_C_FORMAT_GUID)
        self.assertEqual('', u)

    def test_valid_reg_input_with_brackets(self):
        """ check the reg_format functions are able to handle extra {} as reg format sometimes has brackets """
        SAMPLE_DATA_REG_FORMAT_GUID_WITH = "{66341ae8-668f-4192-b44d-5f87b868f041}"
        SAMPLE_DATA_REG_FORMAT_GUID = "66341ae8-668f-4192-b44d-5f87b868f041"
        u = GuidParser.uuid_from_guidstring(SAMPLE_DATA_REG_FORMAT_GUID_WITH)
        self.assertEqual(SAMPLE_DATA_REG_FORMAT_GUID, GuidParser.reg_guid_str_from_uuid(u))

    def test_valid_reg_input_with_spaces(self):
        """ check the reg_format functions are able to handle extra spaces """
        SAMPLE_DATA_REG_FORMAT_GUID_WITH = "    66341ae8-668f-4192-b44d-5f87b868f041           "
        SAMPLE_DATA_REG_FORMAT_GUID = "66341ae8-668f-4192-b44d-5f87b868f041"
        u = GuidParser.uuid_from_guidstring(SAMPLE_DATA_REG_FORMAT_GUID_WITH)
        self.assertEqual(SAMPLE_DATA_REG_FORMAT_GUID, GuidParser.reg_guid_str_from_uuid(u))

    def test_valid_c_format_input_with_spaces(self):
        """ check the c_format functions are able to handle extra spaces """
        SAMPLE_DATA_C_FORMAT_GUID = "   {  0x66341ae8, 0x668f, 0x4192, {0xb4, 0x4d, 0x5f, 0x87, 0xb8, 0x68, 0xf0, 0x41  }  }   "  # noqa: E501
        SAMPLE_DATA_REG_FORMAT_GUID = "66341ae8-668f-4192-b44d-5f87b868f041"
        u = GuidParser.uuid_from_guidstring(SAMPLE_DATA_C_FORMAT_GUID)
        self.assertEqual(SAMPLE_DATA_REG_FORMAT_GUID, GuidParser.reg_guid_str_from_uuid(u))



# @file guid_list_test.py
# Contains unit test routines for the guid list class.
#
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import unittest
import os
import tempfile
import shutil
from edk2toollib.uefi.edk2.guid_list import GuidList, GuidListEntry


class TestGuidListEntry(unittest.TestCase):

    def test_valid_input(self):
        GUID = "66341ae8-668f-4192-b44d-5f87b868f041"
        NAME = "testguid"
        FILEPATH = os.path.dirname(os.path.abspath(__file__))
        t = GuidListEntry(NAME, GUID, FILEPATH)
        self.assertEqual(t.name, NAME)
        self.assertEqual(t.guid, GUID)
        self.assertEqual(t.absfilepath, FILEPATH)
        self.assertTrue(GUID in str(t))
        self.assertTrue(FILEPATH in str(t))
        self.assertTrue(NAME in str(t))


class TestGuidList(unittest.TestCase):

    SAMPLE_DEC_FILE = \
        """## @file
TestDecFile
##

[Defines]
  DEC_SPECIFICATION              = 0x00010005
  PACKAGE_NAME                   = TestDecParserPkg
  PACKAGE_UNI_FILE               = TestDecParserPkg.uni
  PACKAGE_GUID                   = 57e8a49e-1b3f-41a0-a552-55ad831c15a8
  PACKAGE_VERSION                = 0.1

[Includes]
  Include

[LibraryClasses]
  ##  @libraryclass  Provide comment for fakelib
  #
  FakeLib|Include/Library/FakeLib.h

[Guids]
  gFakeTokenSpace =  {0x7c004f69, 0xd730, 0x4904, {0x83, 0xe1, 0x4b, 0xbf, 0x39, 0x3a, 0xfc, 0xb1}}
  gFake2TokenSpace = {0x7c004f69, 0xd730, 0x4904, {0x83, 0xe1, 0x4b, 0xbf, 0x39, 0x3a, 0xfc, 0xb2}}
  gFakeT3okenSpace = {0x7c004f69, 0xd730, 0x4904, {0x83, 0xe1, 0x4b, 0xbf, 0x39, 0x3a, 0xfc, 0xb3}}

[Protocols]
  ## None
  gFakeProtocol     = {0xe63e2ccd, 0x786e, 0x4754, {0x96, 0xf8, 0x5e, 0x88, 0xa3, 0xf0, 0xaf, 0x85}}

[Ppis]
  gFakePpi  =  {0xeeef868e, 0x5bf5, 0x4e48, {0x92, 0xa1, 0xd7, 0x6e, 0x02, 0xe5, 0xb9, 0xa7}}
  gFake2Ppi =  {0xeeef868e, 0x5bf5, 0x4e48, {0x92, 0xa1, 0xd7, 0x6e, 0x02, 0xe5, 0xb9, 0xa8}}

[UserExtensions.TianoCore."ExtraFiles"]
  PcAtChipsetPkgExtra.uni
"""

    SAMPLE_INF_FILE = \
        """## @file
#  Sample UEFI Application Reference EDKII Module.
#
#  This is a sample shell application that will print "UEFI Hello World!" to the
#  UEFI Console based on PCD setting.
#
#  It demos how to use EDKII PCD mechanism to make code more flexible.
#
#  Copyright (c) 2008 - 2018, Intel Corporation. All rights reserved.<BR>
#
#  SPDX-License-Identifier: BSD-2-Clause-Patent
#
#
##

[Defines]
  INF_VERSION                    = 0x00010005
  BASE_NAME                      = HelloWorld
  MODULE_UNI_FILE                = HelloWorld.uni
  FILE_GUID                      = 6987936E-ED34-44db-AE97-1FA5E4ED2116
  MODULE_TYPE                    = UEFI_APPLICATION
  VERSION_STRING                 = 1.0
  ENTRY_POINT                    = UefiMain

#
#  This flag specifies whether HII resource section is generated into PE image.
#
  UEFI_HII_RESOURCE_SECTION      = TRUE

#
# The following information is for reference only and not required by the build tools.
#
#  VALID_ARCHITECTURES           = IA32 X64 EBC
#

[Sources]
  HelloWorld.c
  HelloWorldStr.uni

[Packages]
  MdePkg/MdePkg.dec
  MdeModulePkg/MdeModulePkg.dec

[LibraryClasses]
  UefiApplicationEntryPoint
  UefiLib
  PcdLib

[FeaturePcd]
  gEfiMdeModulePkgTokenSpaceGuid.PcdHelloWorldPrintEnable   ## CONSUMES

[Pcd]
  gEfiMdeModulePkgTokenSpaceGuid.PcdHelloWorldPrintString   ## SOMETIMES_CONSUMES
  gEfiMdeModulePkgTokenSpaceGuid.PcdHelloWorldPrintTimes    ## SOMETIMES_CONSUMES

[UserExtensions.TianoCore."ExtraFiles"]
  HelloWorldExtra.uni
"""

    def __init__(self, *args, **kwargs):
        self.test_dir = None
        super().__init__(*args, **kwargs)

    def prep_workspace(self):
        self.clean_workspace()
        self.test_dir = tempfile.mkdtemp()

    def clean_workspace(self):
        if self.test_dir is None:
            return
        if os.path.isdir(self.test_dir):
            shutil.rmtree(self.test_dir)
            self.test_dir = None

    def setUp(self):
        self.prep_workspace()

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def tearDown(self):
        self.clean_workspace()

    def test_valid_input(self):
        # write dec and inf sample files to tempdir
        with open(os.path.join(self.test_dir, "test.dec"), "w") as f:
            f.write(TestGuidList.SAMPLE_DEC_FILE)
        with open(os.path.join(self.test_dir, "test.inf"), "w") as f:
            f.write(TestGuidList.SAMPLE_INF_FILE)

        ResultList = GuidList.guidlist_from_filesystem(self.test_dir)
        self.assertEqual(len(ResultList), 8)

    def test_valid_input_recursive_filesystem(self):
        # write dec and inf sample files to tempdir
        os.makedirs(os.path.join(self.test_dir, "Build"))
        with open(os.path.join(self.test_dir, "Build", "test.dec"), "w") as f:
            f.write(TestGuidList.SAMPLE_DEC_FILE)
        with open(os.path.join(self.test_dir, "test.inf"), "w") as f:
            f.write(TestGuidList.SAMPLE_INF_FILE)

        ResultList = GuidList.guidlist_from_filesystem(self.test_dir)
        self.assertEqual(len(ResultList), 8)

    def test_incomplete_inf_contents(self):
        # write dec and inf sample files to tempdir
        with open(os.path.join(self.test_dir, "test.dec"), "w") as f:
            f.write(TestGuidList.SAMPLE_DEC_FILE)
        with open(os.path.join(self.test_dir, "test.inf"), "w") as f:
            f.write(TestGuidList.SAMPLE_INF_FILE.replace("BASE_NAME", "INVALID"))

        ResultList = GuidList.guidlist_from_filesystem(self.test_dir)
        self.assertEqual(len(ResultList), 7)

    def test_incomplete_dec_contents(self):
        # write dec and inf sample files to tempdir
        with open(os.path.join(self.test_dir, "test.dec"), "w") as f:
            f.write(TestGuidList.SAMPLE_DEC_FILE.replace("PACKAGE_NAME", "GARBAGE"))
        with open(os.path.join(self.test_dir, "test.inf"), "w") as f:
            f.write(TestGuidList.SAMPLE_INF_FILE)

        ResultList = GuidList.guidlist_from_filesystem(self.test_dir)
        self.assertEqual(len(ResultList), 7)

    def test_unsupported_file_type(self):
        # write dec and inf sample files to tempdir
        with open(os.path.join(self.test_dir, "test.unsupported_dec"), "w") as f:
            f.write(TestGuidList.SAMPLE_DEC_FILE)
        with open(os.path.join(self.test_dir, "test.unsupported_inf"), "w") as f:
            f.write(TestGuidList.SAMPLE_INF_FILE)

        ResultList = GuidList.guidlist_from_filesystem(self.test_dir)
        self.assertEqual(len(ResultList), 0)

    def test_ignore_file_pattern(self):
        # write dec and inf sample files to tempdir
        with open(os.path.join(self.test_dir, "test.dec"), "w") as f:
            f.write(TestGuidList.SAMPLE_DEC_FILE)
        with open(os.path.join(self.test_dir, "test.inf"), "w") as f:
            f.write(TestGuidList.SAMPLE_INF_FILE)

        ResultList = GuidList.guidlist_from_filesystem(self.test_dir, ["*.inf"])
        self.assertEqual(len(ResultList), 7)

    def test_ignore_folder_pattern(self):
        # write dec and inf sample files to tempdir
        os.makedirs(os.path.join(self.test_dir, "my_ignore"))
        with open(os.path.join(self.test_dir, "my_ignore", "test.dec"), "w") as f:
            f.write(TestGuidList.SAMPLE_DEC_FILE)
        with open(os.path.join(self.test_dir, "test.inf"), "w") as f:
            f.write(TestGuidList.SAMPLE_INF_FILE)

        ResultList = GuidList.guidlist_from_filesystem(self.test_dir, ["my_ignore"])
        self.assertEqual(len(ResultList), 1)

    def test_ignore_file_pattern2(self):
        # write dec and inf sample files to tempdir
        os.makedirs(os.path.join(self.test_dir, "my_ignore"))
        with open(os.path.join(self.test_dir, "my_ignore", "test.dec"), "w") as f:
            f.write(TestGuidList.SAMPLE_DEC_FILE)
        with open(os.path.join(self.test_dir, "test.inf"), "w") as f:
            f.write(TestGuidList.SAMPLE_INF_FILE)

        ResultList = GuidList.guidlist_from_filesystem(self.test_dir, ["my_ignore/test.dec"])
        self.assertEqual(len(ResultList), 1)



## @file
# Module that encodes and decodes a EFI_FIRMWARE_IMAGE_AUTHENTICATION with
# certificate data and payload data.
#
# Copyright (c) 2018 - 2019, Intel Corporation. All rights reserved.<BR>
# Copyright (c) Microsoft Corporation
# SPDX-License-Identifier: BSD-2-Clause-Patent
#

"""Module for encoding and decoding EFI_FIRMWARE_IMAGE_AUTHENTICATION with certificate data and payload data."""

import struct

from edk2toollib.uefi.wincert import WinCertUefiGuid
from edk2toollib.uefi.edk2.fmp_payload_header import FmpPayloadHeaderClass


class FmpAuthHeaderClass (object):
    r"""An object representing an EFI_FIRMWARE_IMAGE_AUTHENTICATION.

    Can parse or produce an EFI_FIRMWARE_IMAGE_AUTHENTICATION structure/byte buffer.

    Attributes:
        MonotonicCount (int):       It is included in the signature of AuthInfo. It is used to ensure freshness/no
                                    replay. It is incremented during each firmware image operation.
        AuthInfo (WinCertUefiGuid): Provides the authorization for the firmware image operations.
        Payload (str):              string representing payload as bytes (i.e. b'\x01\x00\x03')
        FmpPayloadHeader (FmpPayloadHeaderClass): Header for the payload

    ```
    typedef struct {
        UINT64                                  MonotonicCount;
        WIN_CERTIFICATE_UEFI_GUID               AuthInfo;
    } EFI_FIRMWARE_IMAGE_AUTHENTICATION;
    ```
    """
    _MonotonicCountFormat = '<Q'
    _MonotonicCountSize = struct.calcsize(_MonotonicCountFormat)

    def __init__(self):
        """Inits an empty object."""
        self.MonotonicCount = 0
        self.AuthInfo = WinCertUefiGuid()
        self.Payload = b''
        self.FmpPayloadHeader = None

    def Encode(self):
        r"""Serializes the Auth header + AuthInfo + Payload/FmpPayloadHeader.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        FmpAuthHeader = struct.pack(
            self._MonotonicCountFormat,
            self.MonotonicCount
        )

        if self.FmpPayloadHeader is not None:
            return FmpAuthHeader + self.AuthInfo.Encode() + self.FmpPayloadHeader.Encode()
        else:
            return FmpAuthHeader + self.AuthInfo.Encode() + self.Payload

    def Decode(self, Buffer):
        """Loads data into the Object by parsing a buffer.

        Args:
            Buffer (obj): Buffer containing the data

        Returns:
            (str): string of binary representing the payload

        Raises:
            (ValueError): Invalid Buffer
        """
        if len(Buffer) < self._MonotonicCountSize:
            raise ValueError
        (MonotonicCount,) = struct.unpack(
            self._MonotonicCountFormat,
            Buffer[:self._MonotonicCountSize]
        )
        self.MonotonicCount = MonotonicCount

        self.Payload = self.AuthInfo.Decode(Buffer[self._MonotonicCountSize:])
        if len(self.Payload) > 0:
            self.FmpPayloadHeader = FmpPayloadHeaderClass()
            self.FmpPayloadHeader.Decode(self.Payload)
        return self.Payload

    def IsSigned(self, Buffer):
        """Parses the buffer and returns if the Cert is signed or not.

        Returns:
            (bool): True if signed
            (bool): False if invalid buffer
            (bool): False if not signed
        """
        if len(Buffer) < self._MonotonicCountSize:
            return False

        auth_info = WinCertUefiGuid(Buffer[self._MonotonicCountSize:])
        if auth_info.CertType != WinCertUefiGuid._EFI_CERT_TYPE_PKCS7_GUID.bytes_le:
            return False
        return True

    def DumpInfo(self):
        """Prints object to console."""
        print('EFI_FIRMWARE_IMAGE_AUTHENTICATION.MonotonicCount                = {MonotonicCount:016X}'
              .format(MonotonicCount=self.MonotonicCount))
        self.AuthInfo.DumpInfo()
        print('sizeof (Payload)                                                = {Size:08X}'
              .format(Size=len(self.Payload)))
        if self.FmpPayloadHeader is not None:
            self.FmpPayloadHeader.DumpInfo()



# @file utility_functions_test.py
# unit test for utility_functions module.
#
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import unittest
import io
import os
import sys
import edk2toollib.utility_functions as utilities


class DesiredClass():
    def __str__(self):
        return "DesiredClass"


class ChildOfDesiredClass(DesiredClass):
    def __str__(self):
        return "Child of DesiredClass"


class GrandChildOfDesiredClass(ChildOfDesiredClass):
    def __str__(self):
        return "GrandChild of DesiredClass"


'''
The current solution can't handle a brother class
class BrotherOfChildOfDesiredClass(DesiredClass):
    def __str__(self):
        return "Brother of Child of DesiredClass"
'''


class UtilityFunctionsTest(unittest.TestCase):

    def test_RunPythonScript(self):  # simple- run yourself!
        path = __file__
        working_dir = os.path.dirname(__file__)
        ret = utilities.RunPythonScript(path, "", working_dir)
        self.assertEqual(ret, 0)
        # test it with all the named parameters
        ret = utilities.RunPythonScript(path, "", capture=False, workingdir=working_dir,
                                        raise_exception_on_nonzero=True)
        self.assertEqual(ret, 0)
        # now try a bad path
        bad_path = __file__ + ".super.bad"
        ret = utilities.RunPythonScript(bad_path, "", capture=False, workingdir=None, raise_exception_on_nonzero=False)
        self.assertNotEqual(ret, 0)
        # now we expect it to throw an exception
        with self.assertRaises(Exception):
            ret = utilities.RunPythonScript(bad_path, "", capture=False, workingdir=None,
                                            raise_exception_on_nonzero=True)
            self.assertNotEqual(ret, 0)

    def test_locate_class_in_module(self):
        module = sys.modules[__name__]

        found_class = utilities.locate_class_in_module(module, DesiredClass)
        self.assertIsNotNone(found_class)
        self.assertEqual(found_class, GrandChildOfDesiredClass)


class HexdumpTest(unittest.TestCase):
    """Tests hexdump"""

    def test_hexdump_basic_usage(self):
        """Basic Usage Test"""
        test = b"Hello UEFI!"
        output = io.StringIO()

        newline = '\n'
        expected_output = f"0x00 - 0x48 0x65 0x6c 0x6c 0x6f 0x20 0x55 0x45 - 0x46 0x49 0x21                          Hello UEFI! {newline}" # noqa
        utilities.hexdump(test, out_fs=output)

        self.assertEqual(expected_output, output.getvalue())

    def test_hexdump_basic_usage_offset_start(self):
        """Basic Usage Test"""
        test = b"Hello UEFI!"
        output = io.StringIO()

        newline = '\n'
        expected_output = f"0x40000000 - 0x48 0x65 0x6c 0x6c 0x6f 0x20 0x55 0x45 - 0x46 0x49 0x21                          Hello UEFI! {newline}" # noqa
        utilities.hexdump(test, offset_start=0x4000_0000, out_fs=output)

        self.assertEqual(expected_output, output.getvalue())

    def test_hexdump_basic_usage_16_bytes(self):
        """Basic 16 byte Test"""

        test = b"0123456789abcdef"
        output = io.StringIO()

        newline = '\n'

        expected_output = f"0x00 - 0x30 0x31 0x32 0x33 0x34 0x35 0x36 0x37 - 0x38 0x39 0x61 0x62 0x63 0x64 0x65 0x66 0123456789abcdef {newline}" # noqa

        utilities.hexdump(test, out_fs=output)

        self.assertEqual(expected_output, output.getvalue())

    def test_hexdump_basic_usage_32_bytes(self):
        """Basic 32 byte Test"""
        test = b"0123456789abcdef0123456789abcdef"
        output = io.StringIO()

        newline = '\n'

        expected_output =  f"0x40000000 - 0x30 0x31 0x32 0x33 0x34 0x35 0x36 0x37 - 0x38 0x39 0x61 0x62 0x63 0x64 0x65 0x66 0123456789abcdef {newline}" # noqa
        expected_output += f"0x40000010 - 0x30 0x31 0x32 0x33 0x34 0x35 0x36 0x37 - 0x38 0x39 0x61 0x62 0x63 0x64 0x65 0x66 0123456789abcdef {newline}" # noqa

        utilities.hexdump(test, offset_start=0x4000_0000, out_fs=output)

        self.assertEqual(expected_output, output.getvalue())

    def test_hexdump_basic_usage_33_bytes(self):
        """Basic 32 byte Test"""
        test = b"0123456789abcdef0123456789abcdef0"
        output = io.StringIO()

        newline = '\n'
        expected_output =  f"0x00 - 0x30 0x31 0x32 0x33 0x34 0x35 0x36 0x37 - 0x38 0x39 0x61 0x62 0x63 0x64 0x65 0x66 0123456789abcdef {newline}" # noqa
        expected_output += f"0x10 - 0x30 0x31 0x32 0x33 0x34 0x35 0x36 0x37 - 0x38 0x39 0x61 0x62 0x63 0x64 0x65 0x66 0123456789abcdef {newline}" # noqa
        expected_output += f"0x20 - 0x30                                                                              0 {newline}" # noqa

        utilities.hexdump(test, out_fs=output)

        self.assertEqual(expected_output, output.getvalue())

    def test_hexdump_basic_usage_32_bytes_and_line_cont(self):
        """Basic line continuation Test"""
        test = b"0123456789abcde\\0123456789abcde\\"
        output = io.StringIO()

        newline = '\n'
        expected_output =  f"0x00 - 0x30 0x31 0x32 0x33 0x34 0x35 0x36 0x37 - 0x38 0x39 0x61 0x62 0x63 0x64 0x65 0x5c 0123456789abcde\\ {newline}" # noqa
        expected_output += f"0x10 - 0x30 0x31 0x32 0x33 0x34 0x35 0x36 0x37 - 0x38 0x39 0x61 0x62 0x63 0x64 0x65 0x5c 0123456789abcde\\ {newline}" # noqa

        utilities.hexdump(test, out_fs=output)

        self.assertEqual(expected_output, output.getvalue())

    def test_hexdump_basic_usage_empty(self):
        """Ensures empty response"""
        test = io.BytesIO(b"")
        output = io.StringIO()
        newline = '\n'

        expected_output = f"{newline}"

        utilities.hexdump(test, out_fs=output)

        self.assertEqual(expected_output, output.getvalue())


class ExportCTypeArrayTest(unittest.TestCase):
    """Tests export_c_type_array"""

    def test_export_c_type_array_basic_usage(self):
        """Basic Usage Test"""
        original_bytes = b"Hello UEFI!"
        length = len(original_bytes)
        test = io.BytesIO(original_bytes)
        output = io.StringIO()

        newline = '\n'

        expected_output =  f"UINT8 TestVariable[{length}] = {{{newline}"  # noqa
        expected_output += f"    0x48, 0x65, 0x6c, 0x6c, 0x6f, 0x20, 0x55, 0x45, 0x46, 0x49, 0x21                                 // Hello UEFI! {newline}"  # noqa
        expected_output += f"}};{newline*2}"

        utilities.export_c_type_array(test, "TestVariable", output)

        self.assertEqual(expected_output, output.getvalue())

    def test_export_c_type_array_include_length_variable(self):
        """Basic Usage Test"""
        original_bytes = b"Hello UEFI!"
        length = len(original_bytes)
        test = io.BytesIO(original_bytes)
        output = io.StringIO()

        newline = '\n'

        expected_output =  f"UINT8 TestVariable[{length}] = {{{newline}"  # noqa
        expected_output += f"    0x48, 0x65, 0x6c, 0x6c, 0x6f, 0x20, 0x55, 0x45, 0x46, 0x49, 0x21                                 // Hello UEFI! {newline}"  # noqa
        expected_output += f"}};{newline*2}"
        expected_output += f"UINTN TestVariableLength = sizeof TestVariable;{newline*2}"

        utilities.export_c_type_array(test, "TestVariable", output, length_variable_name="TestVariableLength")

        self.assertEqual(expected_output, output.getvalue())

    def test_export_c_type_array_16_bytes(self):
        """Basic 16 byte Test"""

        original_bytes = b"0123456789abcdef"
        length = len(original_bytes)
        test = io.BytesIO(original_bytes)
        output = io.StringIO()

        newline = '\n'

        expected_output =  f"UINT8 TestVariable[{length}] = {{{newline}"  # noqa
        expected_output += f"    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66   // 0123456789abcdef {newline}"  # noqa
        expected_output += f"}};{newline*2}"

        utilities.export_c_type_array(test, "TestVariable", output)

        self.assertEqual(expected_output, output.getvalue())

    def test_export_c_type_array_32_bytes(self):
        """Basic 32 byte Test"""
        original_bytes = b"0123456789abcdef0123456789abcdef"
        length = len(original_bytes)
        test = io.BytesIO(original_bytes)
        output = io.StringIO()

        newline = '\n'

        expected_output =  f"UINT8 TestVariable[{length}] = {{{newline}"  # noqa
        expected_output += f"    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66,  // 0123456789abcdef {newline}"  # noqa
        expected_output += f"    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66   // 0123456789abcdef {newline}"  # noqa
        expected_output += f"}};{newline*2}"

        utilities.export_c_type_array(test, "TestVariable", output)

        self.assertEqual(expected_output, output.getvalue())

    def test_export_c_type_array_33_bytes(self):
        """Basic 32 byte Test"""
        original_bytes = b"0123456789abcdef0123456789abcdef0"
        length = len(original_bytes)
        test = io.BytesIO(original_bytes)
        output = io.StringIO()

        newline = '\n'

        expected_output =  f"UINT8 TestVariable[{length}] = {{{newline}"  # noqa
        expected_output += f"    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66,  // 0123456789abcdef {newline}"  # noqa
        expected_output += f"    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66,  // 0123456789abcdef {newline}"  # noqa
        expected_output += f"    0x30                                                                                             // 0 {newline}"  # noqa
        expected_output += f"}};{newline*2}"

        utilities.export_c_type_array(test, "TestVariable", output)

        self.assertEqual(expected_output, output.getvalue())

    def test_export_c_type_array_32_bytes_and_line_cont(self):
        """Basic line continuation Test"""
        original_bytes = b"0123456789abcde\\0123456789abcde\\"
        length = len(original_bytes)
        test = io.BytesIO(original_bytes)
        output = io.StringIO()

        newline = '\n'

        expected_output =  f"UINT8 TestVariable[{length}] = {{{newline}"  # noqa
        expected_output += f"    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x61, 0x62, 0x63, 0x64, 0x65, 0x5c,  // 0123456789abcde\\ {newline}"  # noqa
        expected_output += f"    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x61, 0x62, 0x63, 0x64, 0x65, 0x5c   // 0123456789abcde\\ {newline}"  # noqa
        expected_output += f"}};{newline*2}"

        utilities.export_c_type_array(test, "TestVariable", output)

        self.assertEqual(expected_output, output.getvalue())

    def test_export_c_type_array_empty(self):
        """Ensure exception is raised if the array is length 0"""
        binary_data = io.BytesIO(b"")
        output = io.StringIO()

        with self.assertRaises(ValueError):
            utilities.export_c_type_array(binary_data, "TestVariable", output)

# DO NOT PUT A MAIN FUNCTION HERE
# this test runs itself to test runpython script, which is a tad bit strange yes.
